﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VisTool {
    [AttributeUsage(AttributeTargets.Field | AttributeTargets.Method | AttributeTargets.Delegate | AttributeTargets.Property | AttributeTargets.Event)]
    public class VisPropAtt : Attribute {
        public bool ReadOnly=false; // Whether the property can be edited in the Property grid.
        public bool Exposed=true;  // Whether the property can be accessed from a formula. False when a common property (in VicComponent) is overridden as non-accessible.
        public bool InPopup = true; // Some properties are Exposed, but don't appear in Popup, e.g. Rows. Not useful, SL 02-08-14. 
        public bool ShowInitially=false; // Whether the property is shown initially in the PropertyGrid
        public string DefaultFormula = null; // The default property formula. Shown also at ReadOnly as a formula comment. 
        // The default value for the property may be set by the constructor or iin the declaration of the internal property, e.g. int top = 5. 
        public string PropertyCategory = "Special"; // The property grid category where the property is shown.
        public int Position=100; // Position of the property in the PropertyGrid within the category. Default: Show them last.
        public enum PropertyType { Property, Method, Eventhandler };
        public PropertyType PropType;
        public string[] Parameters; // A list of the parameters. Intended for IntelliSense.

        public string LongHelp = null; // Should contain an example and maybe an URL. When null, the ShortHelp is shown.
        public string ShortHelp = "There is no tooltip for this property";

        public VisPropAtt(PropertyType propType, string defaultFormula) {
            this.PropType = propType;
            this.DefaultFormula = defaultFormula;
        }
    }

    [AttributeUsage(AttributeTargets.Class | AttributeTargets.Struct)]
    public class VisClassAtt : Attribute {
        public bool AutoInstances = true; // Whether the kernel generates the instances or the component does it. 
        public string ShortHelp = "There is no explanation for this class"; // Should contain an example and maybe an URL.
        public string LongHelp = null;
        public string ToolboxCategory = "Control"; // Allows Toolbox to show componenents by class.
        public string Icon = ""; // Used in Toolbox
        // E.g. a combobox generates its own dropdown list. A traditional data grid generates its own table rows.
        public bool ShowInToolbox = true; // Most component classes are shown. The base class ViisCompoenent is not, nor is PanelCell.

        public VisClassAtt() { }
    }

    [AttributeUsage(AttributeTargets.Field | AttributeTargets.Method | AttributeTargets.Delegate | AttributeTargets.Property | AttributeTargets.Event)]
    public class VisSystemAtt : Attribute {
        public enum ItemType { SystemMethod, Statement, Field, UvisEnum };
        public bool Exposed=true;  // Whether the System item can be shown in popup.
        public bool SQL = false; // Whether the method is allowed in an SQL-string, i.e. passed on to the database engine. 
        // When not allowed, the method will be evaluated and inserted as a constant when the Rows formula is evaluated.
        public int Position=0; // property position in popup
        public ItemType itemType;
        public string[] Parameters; // A list of the parameters. Used by IntelliSense.

        public string LongHelp = "There is no explanation for this item"; // Should contain an example and maybe an URL.
        public string ShortHelp = "There is no tooltip for this item";

        public VisSystemAtt(ItemType type) {
            this.itemType = type;
        }
    }
}
