﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Threading;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using VisTool;

namespace VisTool
{
    public partial class Parser {
        // At the moment this index is not really used, since the only datamap we use will always be VisTool.Datamaps[0]...
        // BUT this parser is able to actually read and "instantiate" any number of datamaps
        int currentDataMapIndex = -1;
        string CurrentDBName;

        // Variables that show the state of parsing tables and relations:
        Token chainStart;
        string fromTableName;
        string relationName;
        string toTableName;
        CardinalityType cardinality = CardinalityType.onemany;
        List<string> fromKeyNames = new List<string>(0);
        List<string> toKeyNames = new List<string>(0);
        Token startJoinSnippet;
        Token startTableSpecs=null;

        // VismFile = Modes { DataBaseSpec { TableSpec { RelationSpec } } }
        // Modes
        // DatabaseSpec = "DATABASE:" connection_string
        // TableSpec = "Table:" name EOL
        // RelationSpec = relation_name ":" [one | many] ", join" tabel_name "on" 
        //                tabel_name "." field_name = name "." field_name 
        //                [EOL | EOF]
        // Although the syntax allows for it, we can have only one database at present.

        public static bool OpenMapFile(string path) { // false after errors that show a message box, e.g. database not found.
            StreamReader str=null;
            Thread.CurrentThread.CurrentCulture = VisEngine.codeCulture; // Vism file will use English number formats and separators.
            try {
                // VisEngine initializes itself through initializers and static constructors. SL 22-04-13

                str = new StreamReader(path);
                Parser parser = new VisTool.Parser(str, false);
                int fileNameStart = path.LastIndexOf("\\")+1;
                VisEngine.ApplicationDirectory = path.Substring(0, fileNameStart-1) + "\\";
                VisEngine.MapFile = path.Substring(fileNameStart, path.LastIndexOf(".") - fileNameStart);

                parser.ReadAllExceptRelations(); // Read Modes, DataBaseSpec, TableSpec except Relations.
                // This is the first pass of reading the file. Here any Alias for a table name is handled.
                parser.Reset(); // Start reading at the first table.
                parser.ReadTablesAndRelations(); // This is the second pass. Relations can now be interpreted according to table Aliases.
                str.Close();
                return true;
            }
            catch (System.Exception ex) {
                MessageBox.Show("Map file \"" + path + "\" " + ex.Message); 
                if (str != null) str.Close(); 
                return false;
            }
        }

        void Reset( ) {
            la = pt = tokens = startTableSpecs;
            //StreamReader str = (StreamReader)rd;
            //str.BaseStream.Position = 0;
            //la = tokens = pt = NextToken();
            //Peek();
        }

        void ReadAllExceptRelations() {
            while (la.kind == TkKind.EOL) Get(); // Skip leading blank lines
            try {
                Modes();
                while (!la.EOF) {
                    // If the database cannot be initialized for some reason, then we do not want to continue, we want to get notified by an error
                    if (!DataBaseSpec()) break;
                    startTableSpecs = la;
                    while (la.val.ToUpper() != "DATABASE" && !la.EOF) { // Avoid endless loop after: -------- \n dirt
                        string tbl_name;
                        TableSpec(out tbl_name, CurrentDBName);
                        JumpSeparatorLine(); // Skip relations.
                    }
                }
            }
            catch (Exception) {
                //error(e.Message, token);
                throw;
            }
        }

        /* 
            * LimitedRows: [true | false], maxRows
            *          bool restricted, int max_rows interval: 0-anything; 0 = one
            * InteractionMode: AsEndUser | AsDesigner | DesignWithControl | DesignWithAlt
            *                  enum
            * SimulatedTime: [true | false], time (, increment)
            *              bool simulatedTime, DateTime simulatedDateTime, float simulatedTimeIncrement (TimeSpan in days)
            * SimulatedUpdate: true | false
            *              bool simulatedUpdate
            * Zoom: number
            * ScreenSize: width, height
            *              int screenWidth, int screenHeight (0, 0 being default)
            * StartUpForm: formName
            *              
        */
        void Modes() {
            string currentMode, modeName;
            while (la.val != "--") { // end when there is a separator line
                chainStart = la;
                if (!VismExpect(TkKind.Ident)) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, null, null, 
                    "Expected a mode name", la); goto Skip; }
                currentMode = token.val.ToUpper(); modeName = token.val;
                if (!VismExpect(":")) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null,  
                    "Expected a colon", la); goto Skip; }

                switch (currentMode) {
                case "LIMITEDROWS": // Formerly "DataView"
                    if (!VismExpect(TkKind.Ident)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", la); goto Skip; }
                    if (token.val.ToUpper() == "TRUE") VisEngine.LimitedRows = true;
                    else if (token.val.ToUpper() == "FALSE") VisEngine.LimitedRows = false;
                    else {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", token); goto Skip; }
                    if (!VismExpect(",")) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a comma", la); goto Skip; }
                    if (!VismExpect(TkKind.Long)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a number", la); goto Skip; }
                    VisEngine.MaxRows = Math.Abs((int)token.obj);
                    break;
                case "INTERACTIONMODE":
                    if (!VismExpect(TkKind.Ident)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"AsDesigner\", \"AsEndUser\", ...", la); goto Skip; }
                    if (token.val.ToUpper() == VisEngine.interactionMode.AsDesigner.ToString().ToUpper())
                        VisEngine.InteractionMode = VisEngine.interactionMode.AsDesigner;
                    else if (token.val.ToUpper() == VisEngine.interactionMode.AsEndUser.ToString().ToUpper())
                        VisEngine.InteractionMode = VisEngine.interactionMode.AsEndUser;
                    else if (token.val.ToUpper() == VisEngine.interactionMode.DesignWithAlt.ToString().ToUpper())
                        VisEngine.InteractionMode = VisEngine.interactionMode.DesignWithAlt;
                    else if (token.val.ToUpper() == VisEngine.interactionMode.DesignWithControl.ToString().ToUpper())
                        VisEngine.InteractionMode = VisEngine.interactionMode.DesignWithControl;
                    else
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                            "Expected \"AsDesigner\", \"AsEndUser\", ...", token);
                    break;
                case "SIMULATEDTIME": 
                    if (!VismExpect(TkKind.Ident)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", la); goto Skip; }
                    if (token.val.ToUpper() == "TRUE") VisEngine.simulatedTime = true;
                    else if (token.val.ToUpper() == "FALSE") VisEngine.simulatedTime = false;
                    else {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", token); goto Skip; }
                    if (!VismExpect(",")) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a comma", la); goto Skip; }
                    if (!VismExpect(TkKind.Time)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a datetime of the form: #28-02-2011 14:38:00#", la); goto Skip; }
                    VisEngine.simulatedDateTime = (DateTime)token.obj;
                    VisEngine.simulatedTimeIncrement = 0;
                    if (la.kind == TkKind.EOL) break; // Optional time increment
                    if (!VismExpect(",")) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a comma", la); goto Skip; }
                    if (la.kind == TkKind.Long) VisEngine.simulatedTimeIncrement = (int)la.obj;
                    else if (la.kind == TkKind.Double) VisEngine.simulatedTimeIncrement = (double)la.obj;
                    else {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a number", la); goto Skip; }
                    Get(); // Accept the number.
                    break;
                case "SIMULATEDUPDATE":
                    if (!VismExpect(TkKind.Ident)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", la); goto Skip; }
                    if (token.val.ToUpper() == "TRUE") VisEngine.simulatedUpdate = true;
                    else if (token.val.ToUpper() == "FALSE") VisEngine.simulatedUpdate = false;
                    else {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected \"true\" or \"false\"", token); goto Skip; }
                    break;
                case "ZOOM":
                    if (!VismExpect(TkKind.Double)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null,
                        "Expected a number", la); goto Skip;
                    }
                    float z = (float)(double)token.obj;
                    if (z > 0) SystemItems.Zoom = z;
                    break;
                case "SCREENSIZE":
                    if (!VismExpect(TkKind.Long)) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a number", la); goto Skip; }
                    VisEngine.screenWidth = (int)token.obj;
                    if (!VismExpect(",")) {
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a comma", la); goto Skip; }
                    if (!VismExpect(TkKind.Long)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a number", la); goto Skip; }
                    VisEngine.screenHeight = (int)token.obj;
                    break;
                case "STARTUPFORM":
                    VisEngine.startUpForm = null; // Indicates no startUpForm
                    if (VismExpect(TkKind.Ident)) {
                        VisEngine.startUpForm = token.val;
                        if (!(la.kind == TkKind.EOL)) ErrorLogging.Log(ErrorMode.CompileTime, Priority.Warning, VisEngine.MapFile, modeName, null, 
                            "Expected the name of a form followed by end-of-line", la);
                    } else if (VismExpect(TkKind.String))
                        ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                            "Write the form name without quotes", la);
                    else ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected the name of a form followed by end-of-line", la);
                    goto Skip;
                default:
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                        "Expected a mode name, e.g. \"LimitedRows\", \"InteractionMode\", ...", token); goto Skip;
                }
                if (!VismExpect(TkKind.EOL)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, modeName, null, 
                    "Expected end-of-line", la); goto Skip; }
                continue;
Skip:
                SkipLine();
            }
            SkipSeparatorLine();
        }

        // DatabaseSpec = "Database:" connection_string
        private bool DataBaseSpec() {
            chainStart = la;
            if (!VismExpect("Database")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.MsgBox, VisEngine.MapFile, null, null, 
                "Expected \"Database:\"", la); SkipLine(); return false; }
            if (!VismExpect(":")) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.MsgBox, VisEngine.MapFile, "Database", null, 
                "Expected a colon", la); SkipLine(); return false; }
            if (!VismExpect(TkKind.Ident)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.MsgBox, VisEngine.MapFile, "Database", null, 
                "Expected a connection string", la); SkipLine(); return false; }
            string connectionString = token.val;
            while (la.kind != TkKind.EOL) {
                Get();
                connectionString += token.space + token.val;
            }
            Get();
            if (!(la.val == "--" || la.val == "-")) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "Database", null, 
                "Expected a separator line (----)", la); SkipLine(); } else SkipSeparatorLine();
            // create a new DataMap in the list in VisEngine and set its connection string
            if (VisEngine.DataMaps.Count > 0) VisEngine.DataMaps[0].Connection.Dispose();
            VisEngine.DataMaps.Clear();
            VisEngine.DataMaps.Add(new DataMap());
            currentDataMapIndex++;
            VisEngine.DataMaps[currentDataMapIndex].ConnectionString = connectionString; // Setting opens the database and loads the Schema. Awful side effect.
            int start1 = connectionString.LastIndexOf("=") + 1;
            int start = connectionString.LastIndexOf("\\", start1);
            if (start < start1) start = start1;
            int length = connectionString.Length - start;
            CurrentDBName = connectionString.Substring(start, length);
            return true;
        }

        // TableSpec = "Table:" name [ AS aliasName ] EOL
        private void TableSpec(out string tbl_name, string db_name) {
            string old_tbl_name = "";
            chainStart = la;
            if (!VismExpect("Table")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, null, 
                "Expected \"Table\"", la); goto Skip; }
            if (!VismExpect(":")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, null, 
                "Expected a colon", la); goto Skip; }
            if (!VismExpect(TkKind.Ident)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, null, 
                "Expected a table name", la); goto Skip; }
            tbl_name = token.val;
            old_tbl_name = tbl_name;
            // find out whether the tbl_name exists as a table in the database
            if (!VisEngine.DataMaps[0].TableExists(tbl_name))
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, "table " + tbl_name, 
                    "The table does not exist in the database", token);
            //check for table alias names
            if (VismExpect("AS") && VismExpect(TkKind.Ident)) {
                tbl_name = token.val;
                if (old_tbl_name.ToUpper() != tbl_name.ToUpper())
                    VisEngine.DataMaps[0].SetTableAliasName(tbl_name, old_tbl_name);
            }

            if (!VismExpect(TkKind.EOL)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, 
                "table " + tbl_name, "Expected end-of-line", la); }
            return;
Skip: SkipLine();
            tbl_name = ""; // Caller will skip to separator line.
        }

        void ReadTablesAndRelations() {
            try {
                //JumpSeparatorLine(); // Skip modes. Done by startTableSpecs.
                //JumpSeparatorLine(); // Skip DatabaseSpec.
                while (!la.EOF) {
                    while (la.val.ToUpper() != "DATABASE" && !la.EOF) { // Avoid endless loop after: -------- \n dirt
                        ReadTable(out fromTableName, CurrentDBName);
                        if (fromTableName == "") {  // if the tbl_name wasn't assigned, we cannot use the relations specified for the table. Skip to separator line.
                            while (la.val != "--" && !la.EOF) {
                                Get();
                            }
                            SkipSeparatorLine();
                        } else {
                            while (la.val != "--" && !la.EOF) {
                                RelationSpec(fromTableName, CurrentDBName);
                            }
                            if (!la.EOF) SkipSeparatorLine();
                            else break;
                        }
                    }
                }
            }
            catch (Exception e) {
                error(e.Message, token);
            }
        }

        // Recognizes the table. It has been created in the first pass.
        private void ReadTable(out string tbl_name, string db_name) {
            chainStart = la; // For error messages.
            if (!VismExpect("Table")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, null, 
                "Expected \"Table\"", la); goto Skip; }
            if (!VismExpect(":")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, null, 
                "Expected a colon", la); goto Skip; }
            if (!VismExpect(TkKind.Ident)) { ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, "Database " + db_name, null, null, 
                "Expected a table name", la); goto Skip; }
            tbl_name = token.val;

            // find out whether the tbl_name exists as a table in the database
            // if (!VisEngine.DataMaps[0].TableExists(tbl_name)) 
            //    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, "Database " + db_name, "table " + tbl_name, null, "The table does not exist in the database", token, chainStart);
            if (VismExpect("AS") && VismExpect(TkKind.Ident))
                tbl_name = token.val;

            if (!VismExpect(TkKind.EOL)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, 
                "table " + tbl_name, "Expected end-of-line", la); }
            return;
Skip:
            SkipLine();
            tbl_name = ""; // Caller will skip to separator line
        }

        // RelationSpec = relationName : [one | many | 1:1 | 1:m | m:1 | m:m] , JOIN toTableName [ AS relationName ] ON tbl1.field1 = . . .
        //e.g. tblMedorder: 1:m, join tblMedOrder on tblPatient.ptID = tblMedOrder.ptID
        private void RelationSpec(string fromTableName, string db_name) {
            chainStart = la;
            fromKeyNames.Clear();
            toKeyNames.Clear();

            if (!VismExpect(TkKind.Ident)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, db_name, 
                "table " + fromTableName, "Expected a relation name", la); SkipLine(); return; }
            relationName = token.val;
            if (!VismExpect(":")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                "relation " + relationName, "Expected a colon", la); SkipLine(); return; }
            if (VisEngine.DataMaps[currentDataMapIndex].GetRelation(fromTableName, relationName) != null) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, "relation " + relationName, 
                    "This relation exists already in table " + fromTableName, token);
                goto Skip;
            }

            //one of the following: "many" and "one" for legacy; 1:m, m:1, m:m or 1:1
            string cardinalityString = "";
            while (la.val != ",") { // Collect "1:m", etc. from several tokens.
                Get();
                cardinalityString += token.val;
            }
            switch (cardinalityString) {
            case "one": cardinality = CardinalityType.manyone; break;
            case "many": cardinality = CardinalityType.onemany; break;
            case "1:1": cardinality = CardinalityType.oneone; break;
            case "1:m": cardinality = CardinalityType.onemany; break;
            case "m:1": cardinality = CardinalityType.manyone; break;
            case "m:m": cardinality = CardinalityType.manymany; break;
            default:
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                    "relation " + relationName, "Expected 1:m, m:m, 1:1 or m:1", token); goto Skip;
            }
            Get(); // Get the ","

            // joinSnippet =  JOIN toTableName [ AS relationName ] ON tbl1.field1 = tbl2.field2 { AND tbl1.field3 = tbl2.field4 }
            // There is no need to limit the join expression. We might use Left(tbl1.field1, 3) = Left(tbl2.field2,3). 
            // However, the field refs have to be collected for inclusion in SELECT lists.

            if (!VismExpect("join")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                    "relation " + relationName, "Expected \"join\"", la); goto Skip;
            }
            if (!VismExpect(TkKind.Ident)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                    "relation " + relationName, "Expected a table name", la); goto Skip;
            }
            toTableName = token.val;
            startJoinSnippet = la;
            if (la.val.ToUpper() == "AS") {
                Get();
                if (!VismExpect(TkKind.Ident) || token.val.ToUpper() != relationName.ToUpper()) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                        "relation " + relationName, "Expected the relation name", la); goto Skip;
                }
            } else
                if (toTableName.ToUpper() != relationName.ToUpper()) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName,
                        "relation " + relationName, "Unknown table name. Use \"AS " + toTableName + "\" to rename the table", token); 
                    goto Skip;
                }

            if (!VismExpect("on")) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                relationName, "Expected \"on\"", la); goto Skip; }

            // syntax: { any | tbl.field } EOL
            while (la.kind != TkKind.EOL) {
                Get();
                if (token.kind == TkKind.Ident && la.val == ".") CheckField();
            }

            string joinSnippet = "join [" + toTableName + "]"; // [ ] needed to to distinguish table name from SQL keywords such as User
            Token t = startJoinSnippet; // The snippet consists of on ...
            while (t != la) {
                joinSnippet += (t.space + t.val);
                t = t.next;
            }
            if (fromKeyNames.Count == 0) ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, 
                "table " + fromTableName, "relation " + relationName, "No fields in " + fromTableName + " are used for the join", token);
            if (toKeyNames.Count == 0) ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, 
                "table " + fromTableName, "relation " + relationName, "No fields in " + relationName + " are used for the join", token);

            Relationship r = new Relationship(relationName, fromTableName, toTableName, cardinality, joinSnippet, fromKeyNames, toKeyNames, VisEngine.DataMaps[currentDataMapIndex]);
            if (!VismExpect(TkKind.EOL)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, 
                "table " + fromTableName, "relation " + relationName, "Expected end-of-line", la); goto Skip; }
            return;
Skip:
            SkipLine();
        }

        private bool CheckField() {
            // Syntax: (fromTableName | relationName) . fieldName
            // Assumes: token = ident, la = "."
            string table = token.val;
            string fieldName;
            Get(); // Get the dot
            if (!VismExpect(TkKind.Ident)) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                    "relation " + relationName, "Expected a field name", la); return false;
            }
            fieldName = token.val;
            if (table.ToUpper() == fromTableName.ToUpper()) {
                if (!VisEngine.DataMaps[currentDataMapIndex].FieldExists(table, fieldName)) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                        "relation " + relationName, fieldName + " is not a field in " + table, la);
                    return false;
                }
                if (!fromKeyNames.Contains(fieldName, StringComparer.OrdinalIgnoreCase)) fromKeyNames.Add(fieldName);
            } else if (table.ToUpper() == relationName.ToUpper()) {
                if (!VisEngine.DataMaps[currentDataMapIndex].FieldExists(toTableName, fieldName)) {
                    ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                        "relation " + relationName, fieldName + " is not a field in " + toTableName, la);
                    return false;
                }
                if (!toKeyNames.Contains(fieldName, StringComparer.OrdinalIgnoreCase)) toKeyNames.Add(fieldName);
            } else {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, "table " + fromTableName, 
                    "relation " + relationName, "Expected " + fromTableName + " or " + relationName + " as table name", la); return false;
            }
            return true;
        }

        void JumpSeparatorLine() {
            chainStart = la;
            while (la.val != "--" && la.val != "-") {
                if (la.EOF)
                    return;
                Get();
            }
            while ((la.val == "--" || la.val == "-") && la.kind != TkKind.EOL) {
                if (la.EOF)
                    return;
                Get();
            }
            if (la.EOF)
                return;
            Get();
        }

        void SkipSeparatorLine() {
            chainStart = la;
            while (la.val == "--" || la.val == "-") Get(); // Read separator line
            if (la.kind != TkKind.EOL) {
                ErrorLogging.Log(ErrorMode.CompileTime, Priority.Error, VisEngine.MapFile, null, null, "Expected a separator line (----)", la);
                while (la.kind != TkKind.EOL) Get(); // Skip the rest of the separator line.
            }
            Get();
        }

        void SkipLine() {
            while (la.kind != TkKind.EOL) Get();
            Get();
        }

        #region EXPECT METHODS
        bool VismExpect(string s) {
            if (s.ToUpper() != la.val.ToUpper()) return false;
            Get(); return true;
        }

        bool VismExpect(TkKind n) {
            if (la.kind != n) return false;
            Get(); return true;
        }

        bool VismExpect(TkKind n, string s) {
            if (la.kind == n && s == la.val.ToUpper()) { Get(); return true; }
            return false;
        }
        #endregion EXPECT METHODS

    }
}
