﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.OleDb;
using System.Data;
using System.IO;
using System.Text.RegularExpressions;
using System.Diagnostics;

namespace VisTool {
    public class Column { // The uVis version of the database schema's Columns.

        public OleDbType ColumnType;
        //public long size;
        //private DataRow schema;
        public string Name;
        public Table Table;

        public string FullName {
            get { return Table.Name + "." + Name; }
        }

        public Column(string name, Table table, DataRow schema) {
            // Was intended also for columns in the query result (RecordSet), but this in not needed anymore. So prefix is useless. SL 16-08-13.
            // Keeps only track of columns in the database tables.
            this.Name = name;
            this.Table = table;
            //this.schema = schema;
            int intType = Int32.Parse("" + schema["DATA_TYPE"]);
            ColumnType = (OleDbType)table.map.getOleDb(intType);
            //if (intType == 130)
            //    size = Int64.Parse("" + schema["CHARACTER_MAXIMUM_LENGTH"]);
            //else if (intType == 3)
            //    size = Int32.MaxValue;
        }

        public string ReadableType {
            get {
                if (ColumnType.ToString() == "WChar")
                    return "String";
                else
                    return ColumnType.ToString();
            }
        }
    }

    public enum CardinalityType { Unknown = 0, onemany, manyone, manymany, oneone }

    public class DataMap {
        Array typeArray = Enum.GetValues(typeof(OleDbType));

        public object getOleDb(int value) {
            OleDbType type = (OleDbType)value;
            //if (value >= 0 && value < typeArray.Length)
            //    return typeArray.GetValue(value);
            //return OleDbType.Error;
            return type;
        }

        public Dictionary<string, Dictionary<string, string>> TableRelationDict = new Dictionary<string, Dictionary<string, string>>();

        string _connectionString;
        static Regex regexRows = new Regex(@"Data\s+Source\s*=\s*(?<matched>[_\w\d]+)\.mdb", RegexOptions.Compiled | RegexOptions.IgnoreCase);

        /// <summary>
        /// Setting the ConnectionString will load database schema
        /// </summary>
        public string ConnectionString {
            get { return this._connectionString; }
            set {
                _connectionString = value;
                if (regexRows.IsMatch(value)) {
                    string rows = regexRows.Match(value).Groups["matched"].Value;
                    if (!string.IsNullOrEmpty(rows) && !System.IO.File.Exists(rows)) {
                        string path = VisEngine.ApplicationDirectory + rows + ".mdb";
                        if (System.IO.File.Exists(path)) this._connectionString = this._connectionString.Replace(rows + ".mdb", path);
                    }
                }
                loadSchema();
            }
        }

        public DataSet DataSet;

        public string GetOldTableName(string newTableName) {
            if (Tables != null && Tables.ContainsKey(newTableName))
                return Tables[newTableName].OldName;
            return newTableName;
        }

        Dictionary<string, DataTable> schema;
        public Dictionary<string, DataTable> Schema {
            get { return schema; }
        }

        Dictionary<string, Table> tables; // Has ignoreCase.
        public Dictionary<string, Table> Tables {
            get { return tables; }
        }

        public static Dictionary<string, string> PKList = new Dictionary<string, string>();

        public void loadSchema() {
            Directory.SetCurrentDirectory(VisEngine.ApplicationDirectory);
            connection = new OleDbConnection(_connectionString);
            schema = new Dictionary<string, DataTable>(StringComparer.InvariantCultureIgnoreCase);
            object[] objArrRestrict = new object[] { null, null, null, "TABLE" };
            DataTable schemaTbl;
            PKList.Clear();

            connection.Open();
            schemaTbl = connection.GetOleDbSchemaTable(OleDbSchemaGuid.Tables, objArrRestrict);

            foreach (DataRow row in schemaTbl.Rows) {
                string tableName = "" + row["TABLE_NAME"];
                objArrRestrict = new object[] { null, null, tableName, null };
                DataTable schemaCols;
                schemaCols = connection.GetOleDbSchemaTable(OleDbSchemaGuid.Columns, objArrRestrict);
                DataTable keySchema = connection.GetOleDbSchemaTable(OleDbSchemaGuid.Primary_Keys, new object[] { null, null, tableName });
                schema.Add(tableName, schemaCols);
                addTable(tableName, schemaCols, keySchema);
                if (keySchema.Rows.Count >0) {
                    PKList.Add(tableName, keySchema.Rows[0].ItemArray[3].ToString());
                }
                else if (!PKList.ContainsKey(tableName))
                        PKList.Add(tableName, "NoPK");
            }
            connection.Close();
        }

        public void SetTableAliasName(string aliasName, string oldName) {
            Table table = null;
            if (tables.ContainsKey(oldName)) {
                table = tables[oldName];
                tables.Remove(oldName);
                tables.Add(aliasName, table);
                table.Name = aliasName;

                if (PKList.ContainsKey(oldName)) {
                    string key = PKList[oldName];
                    PKList.Remove(oldName);
                    PKList.Add(aliasName, key);
                }
            }
        }


        OleDbConnection connection;
        public OleDbConnection Connection {
            get { return connection; }
        }


        public List<Relationship> GetRelations(string fromTableName, string toTableName) {
            List<Relationship> rels = new List<Relationship>();
            Table table = getTable(fromTableName);
            if (table == null) return rels;
            if (table.Relationships == null) return rels;

            Dictionary<string, Relationship>.Enumerator e = table.Relationships.GetEnumerator();
            while (e.MoveNext()) {
                if (e.Current.Value.ToTableName.ToUpperInvariant() == toTableName.ToUpperInvariant()) {
                    rels.Add(e.Current.Value);
                }
            }

            return rels;
        }

        public Relationship GetRelation(string fromTableName, string relationshipName) {
            Table table = getTable(fromTableName);
            if (table != null)
                return table.getRelationship(relationshipName);
            return null;
        }


        public List<Relationship> GetRelations(List<string> fromTableNames, string relName) {
            List<Relationship> rels = new List<Relationship>();
            foreach (string tableName in fromTableNames) {
                Table table = getTable(tableName);
                Dictionary<string, Relationship>.Enumerator e = table.Relationships.GetEnumerator();
                while (e.MoveNext()) {
                    //if (e.Current.Value.ToTableName.ToUpperInvariant() == toTableName.ToUpperInvariant()
                    if (e.Current.Value.RelationName.ToUpperInvariant() == relName.ToUpperInvariant()) {
                        rels.Add(e.Current.Value);
                    }
                }
            }
            return rels;
        }

        public DataMap() {
            DataSet = new DataSet("DataSet");
        }

        public void addTable(Table table) {
            if (tables == null)
                tables = new Dictionary<string, Table>(StringComparer.InvariantCultureIgnoreCase);
            if (tables.ContainsKey(table.Name))
                table = tables[table.Name];
            else
                tables.Add(table.Name, table);
        }

        public void addTable(string tableName, DataTable tableSchema, DataTable keySchema) {
            Table table = null;
            if (tables == null)
                tables = new Dictionary<string, Table>(StringComparer.InvariantCultureIgnoreCase);
            if (tables.ContainsKey(tableName))
                table = tables[tableName];
            else {
                table = new Table(tableName, tableSchema, keySchema, this);
                tables.Add(table.Name, table);
            }
        }

        public string Print() { // For testing ??
            StringBuilder sb = new StringBuilder();
            //the line below is moved to VisEngine, so it does not relate directly to this datamap alone
            //sb.AppendLine("StartForm:" + this.StartUpForm); 
            sb.AppendLine("Database connection:" + this.ConnectionString);
            if (tables != null) {
                foreach (Table table in tables.Values) {
                    sb.Append(Environment.NewLine + Environment.NewLine +
                            "Table" + table.Name + " has the following relationships: " + Environment.NewLine);

                    foreach (Relationship relationship in table.Relationships.Values) {
                        sb.Append("Relationship Name" + relationship.RelationName + Environment.NewLine);
                        foreach (Column column in relationship.FromKeys)
                            sb.Append("From keys" + column.Name + Environment.NewLine);
                        foreach (Column column in relationship.ToKeys)
                            sb.Append("To keys" + column.Name + Environment.NewLine);
                    }
                }
            }
            return sb.ToString();
        }

        public Relationship getRelationship(string relationshipName, string tableName) {
            Table table = null;
            try { table = tables[tableName]; }
            catch (KeyNotFoundException) { throw new System.Exception("Table \""+ tableName + "\" doesn't exist in the map"); }

            return table.getRelationship(relationshipName);
        }

        public Relationship getRelationship(string relationshipName, Table table) {

            try { table = tables[table.Name]; }
            catch (KeyNotFoundException) { throw new System.Exception("Table \"" + table.Name + "\" doesn't exist in the map"); }

            return table.getRelationship(relationshipName);
        }

        public Table getTable(string tableName) {
            if (tables == null)
                throw new System.Exception("No table has been defined");

            try { return tables[tableName]; }
            catch (KeyNotFoundException) { throw new System.Exception("Table " + tableName + " doesn't exist in the map"); }
        }

        public bool FieldExists(string tableName, string columnName) {
            if (tables != null)
                if (tables.ContainsKey(tableName))
                    if ((tables[tableName].Columns != null && tables[tableName].Columns.ContainsKey(columnName)))
                        return true;
            return false;
        }

        public bool FieldExists(string tableName, Column column) {
            return FieldExists(tableName, column.Name);
        }

        public bool FieldExists(Table table, Column column) {
            return FieldExists(table.Name, column.Name);
        }


        public bool TableExists(string tableName) {
            if (tables == null)
                return false;
            return tables.ContainsKey(tableName);
        }

        public bool TableExists(Table table) {
            return TableExists(table.Name);
        }


        public List<string> getTableNames() {
            if (tables != null)
                return tables.Keys.ToList();
            return null;
        }

        public Dictionary<string, List<string>> getColumnNames() {
            Dictionary<string, List<string>> columnNames = null;

            if (tables != null)
                foreach (Table table in tables.Values)
                    if (table.getColumns() != null)
                        columnNames.Add(table.Name, table.getColumns());
            return columnNames;
        }

        public int getRelationshipCount(List<string> tableNames, string relationshipName) {
            int count = 0;
            foreach (string tableName in tableNames) {
                if (tables.ContainsKey(tableName))
                    if (tables[tableName].relationshipExists(relationshipName))
                        count++;
            }
            return count;
        }

        public Relationship getFirstRelationship(List<string> tableNames, string relationshipName) {
            foreach (string tableName in tableNames) {
                if (tables.ContainsKey(tableName)) {
                    Table table = tables[tableName];
                    if (table.relationshipExists(relationshipName))
                        return table.getRelationship(relationshipName);
                }
            }
            return null;
        }

    }

    public struct Field {
        public string prefix; // Relation name, i.e. table name for fields in start of path.
        public string tableName;  // The source table.
        public string fieldName; // Simple field name.
        public string alias; // For aggregate fields.
        public bool isAggregate;

        public override string ToString() { 
            // RecordSetExpr.GetSelect doesn't use this overload anymore, but some other parts do. SL 01-12-13.
            if (!isAggregate) {
                if (!string.IsNullOrEmpty(alias)) 
                    return "(" + prefix + "." + fieldName + ") AS " + alias;
                else return prefix + "." + fieldName;
                //if (!string.IsNullOrEmpty(alias)) return "([" + prefix + "." + field + "]) AS " + alias;
                //else return "[" + prefix + "." + field + "]"; // Okay in select, but causes strange join strings.
            } else
                return fieldName + " AS " + alias;
        }
    }

    public class Relationship {
        public CardinalityType Cardinality;
        public string RelationName; // Relation name. Unique within the FromTable.
        //public string FullRelationName; // FromTableName.RelationName. Used as part of ADO-link ID.
        private DataMap map;
        public string JoinSnippet;  // E.g. JOIN tbl2 ON tbl1.f1 = tbl2.f2 AND tbl1.f3 =tbl2.f4.
        // Used only in RelationExpr where it is either inserted as is.

        public Table FromTable; // Used only between VisTool Table and VisTool Relationship.
        public Table ToTable; // 
        public string FromTableName; // Used in many places.
        public string ToTableName; // Used in PathEngine and more.
        public List<Field> FromKeyFields = new List<Field>(); // The keys in the FromTable, i.e. foreign key fields if m:1, primary key fields if 1:m.
        public List<Field> ToKeyFields = new List<Field>(); // The keys in the ToTable, i.e. primary key fields if m:1, foreign key fields if 1:m.
        public List<Column> FromKeys = new List<Column>(); // Used only in RelationExpr.
        public List<Column> ToKeys = new List<Column>(); // Used only inside RelationsShip. Delete?

        public Relationship(string relationName, string fromTableName, string toTableName, CardinalityType cardinality, string joinSnippet,
            List<string> fromKeyNames, List<string> toKeyNames, DataMap map) {
            // The caller is VismParser. Assumes that the caller has checked the existence of tables and fields and given proper user messages if not.
            // Also assumes that caller has checked that table names and relation names are unique.

            this.map = map;
            this.RelationName = relationName;
            this.FromTableName = fromTableName;
            this.ToTableName = toTableName;
            //this.FullRelationName = FromTableName + "." + RelationName;
            this.Cardinality = cardinality;
            this.JoinSnippet = joinSnippet;
            this.JoinSnippet = joinSnippet;

            FromTable = map.getTable(fromTableName);
            FromTable.addRelationship(this);
            ToTable = map.getTable(toTableName);
            foreach (string s in fromKeyNames) {
                FromKeyFields.Add(new Field { prefix = fromTableName, tableName = toTableName, fieldName = s });
                Column c = FromTable.getColumn(s);
                FromKeys.Add(c);
            }
            foreach (string s in toKeyNames) {
                ToKeyFields.Add(new Field { prefix = toTableName, tableName=toTableName, fieldName = s });
                Column c = ToTable.getColumn(s);
                ToKeys.Add(c);
            }
        }
    }

    public class Table { // VisTool wrapper for System.Data.DataTable. SL 04-02-14
        public string Name;
        public string OldName;
        //private DataTable schema;
        //private DataTable keySchema;
        public DataMap map;

        public List<string> DeleteCommands = new List<string>();
        public Dictionary<string, Relationship> Relationships = new Dictionary<string, Relationship>(StringComparer.InvariantCultureIgnoreCase); // Ignores case.
        public Dictionary<string, Column> Columns = new Dictionary<string, Column>(StringComparer.InvariantCultureIgnoreCase); // Ignores case.
        public Dictionary<string, Column> Keys = new Dictionary<string, Column>(StringComparer.InvariantCultureIgnoreCase);
        public List<Field> KeyList = new List<Field>();

        public Table(string name, DataTable schema, DataTable keySchema, DataMap map) {
            this.Name = name;
            this.OldName = name;
            //this.schema = schema;
            //this.keySchema = keySchema;
            this.map = map;
            foreach (DataRow row in schema.Rows) { // Load all columns.
                string columnName = "" + row["COLUMN_NAME"];
                Column column = new Column(columnName, this, row);
                Columns.Add(column.Name, column);
            }
            foreach (DataRow row in keySchema.Rows) { // Load key columns.
                string columnName = "" + row["COLUMN_NAME"];
                Column column = getColumn(columnName);
                Keys.Add(column.Name, column);
                KeyList.Add(new Field { prefix = Name, tableName = Name, fieldName = column.Name });
            }
        }

        public void addRelationship(Relationship relationship) {
            if (Relationships.ContainsKey(relationship.RelationName))
                relationship = Relationships[relationship.RelationName];
            else {
                Relationships.Add(relationship.RelationName, relationship);
                relationship.FromTable = this;
            }
        }

        public Relationship getRelationship(string relationshipName) {
            if (Relationships != null)
                if (Relationships.ContainsKey(relationshipName))
                    return Relationships[relationshipName];
            return null; // relationship does not exist
        }

        public Column getColumn(string columnName) {
            if (Columns != null)
                if (Columns.ContainsKey(columnName))
                    return Columns[columnName];
            return null; //Column does not exist
        }

        public Column getColumn(Column column) {
            return getColumn(column.Name);
        }

        public List<string> getColumns() {
            if (Columns != null)
                return Columns.Keys.ToList();
            return null;
        }

        public int RelationshipCount() {
            if (Relationships != null)
                return Relationships.Count;
            return 0;
        }

        public bool hasRelationships() {
            return Relationships == null ? false : true;
        }

        public bool relationshipExists(string relationshipName) {
            if (Relationships != null)
                return Relationships.ContainsKey(relationshipName);

            return false;
        }

        private bool isNumber(OleDbType type) {
            return (type == OleDbType.BigInt || type == OleDbType.Binary || type == OleDbType.Boolean || type == OleDbType.Decimal || type == OleDbType.Double || type == OleDbType.Integer || type == OleDbType.LongVarBinary || type == OleDbType.Numeric || type == OleDbType.Single || type == OleDbType.SmallInt || type == OleDbType.TinyInt || type == OleDbType.UnsignedBigInt || type == OleDbType.UnsignedInt || type == OleDbType.UnsignedSmallInt || type == OleDbType.UnsignedTinyInt || type == OleDbType.VarBinary || type == OleDbType.VarNumeric);
        }

        //private bool isPrimaryByFullName(string columnFullName) {
        //    foreach (Column primaryColumn in Keys.Values)
        //        if (String.Compare(columnFullName, primaryColumn.FullName, true) == 0)
        //            return true;
        //    return false;
        //}

        public void FireDeleteCommands(Template tpl) {
            this.map.Connection.Open();
            try {
            foreach (string statement in DeleteCommands) {
                //Debug.Print(statement);
                OleDbCommand deleteCommand = new OleDbCommand(statement, this.map.Connection);
                deleteCommand.ExecuteNonQuery();
            }
            this.map.Connection.Close();
            DeleteCommands.Clear();
            }
            catch (Exception e) { // E.g. referential integrity violated.
                VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp, "DeleteRow: " + e.Message, null, 0);
                this.map.Connection.Close();
                DeleteCommands.Clear();
                throw new LoggedException();
            }
        }

        public string PrepareDeleteCommand(DataRow row, Template tpl) { 
            // When the return value is a string, the row couldn't be deleted and the string is the error message.
            // The row may not have been committed yet. In that case, don't update the database. The caller will delete the row. SL 22-02-14.
            if (row.RowState == DataRowState.Added) return null;
            if (row != null && Keys != null && Keys.Count > 0) {
                string statement = statement = "DELETE FROM " + Name + " ";
                string whereStatement = " WHERE ";
                int i = 0; // Key counter
                foreach (var keyPair in Keys) {
                    Column primaryColumn = keyPair.Value;
                    if (primaryColumn == null) return "Cannot be deleted because the row has no primary key.";
                        string key = row[keyPair.Key].ToString();
                        // When the primary key is Autonumber, the key in rows is null. A requery is needed to get the actual key.
                        // In this case, we cannot delete the row from the database and we shouldn't delete it from the ADO row. 
                        if (String.IsNullOrEmpty(key)) return null; // The caller will delete the ADO row.
                            // The database row cannot be deleted because its primary key has not yet been retrieved from the database
                        whereStatement += primaryColumn.Name + "=" + key;
                        if (i < Keys.Count - 1)
                            whereStatement += " AND ";
                        i++;
                    }
                statement += whereStatement;
                DeleteCommands.Add(statement);
            }
            return null;
        }

        public void FireUpdateCommand(DataRow row, Template tpl, string prefix) { // "this" is the table to be updated.
            // Restructure with StringBuilders and field indexes as in "Retrieve auto keys" below ??
            if (row != null && Keys != null && Keys.Count > 0) {
                string statement = "UPDATE " + Name + "  SET ";
                string whereStatement = " WHERE ";
                bool somethingToSet = false;

                int keyCount = 0; // key counter

                for (int i = 0; i < tpl.RowTpl.RowsExpr.fields.Count; i++) {
                    Field field = tpl.RowTpl.RowsExpr.fields[i];
                    if (field.prefix == prefix) {
                        Column column = getColumn(field.fieldName);
                        if (Keys.ContainsKey(field.fieldName)) { // Primary key, generate Where clause
                            if (row[i] is DBNull)
                                continue; // Cannot update without a primary key.
                            if (keyCount > 1)
                                whereStatement += " AND ";
                            whereStatement += field.fieldName;
                            if (isNumber(column.ColumnType))
                                whereStatement += "=" + row[i];
                            else
                                whereStatement += "='" + row[i] + "'";
                            keyCount++;
                        } else { // Other field: Generate assignment.
                            if (!row[i].Equals(row[i, DataRowVersion.Original])) {
                                // Changed field. Generate assignment field = value
                                statement += field.fieldName; // We update the raw table. Prefixes are not allowed.
                                if (row[i] is DBNull)
                                    statement += "= null, ";
                                else if (isNumber(column.ColumnType))
                                    statement += "= " + Convert.ToString(row[i], VisEngine.codeCulture) + ", ";
                                //statement += "=" + row[dataColumn] + ", ";
                                else if (row[i] is DateTime)
                                    // CodeCulture format (UK) for date is dd/MM/yyyy. This isn't accepted by the database (Access), which requires MM/dd/yyyy,
                                    // but also accepts MM-dd-yyyy. The ToString below generates MM/dd/yyyy.
                                    // Without codeCulture, ToString uses the local setting (e.g. - ) instead of the slashes. 
                                    statement += "= #" + ((DateTime)row[i]).ToString("MM/dd/yy H:mm:ss", VisEngine.codeCulture) + "#, ";
                                else // String. Replace single quotes with double quotes.
                                    statement += "=\"" + row[i].ToString().Replace("\"", "\"\"") + "\", "; 
                                somethingToSet = true;
                            }
                        }
                    }
                }

                if (somethingToSet) {
                    try {
                        if (keyCount != Keys.Count)
                            throw new Exception(Keys.Count + " primary key fields are needed. Only " + keyCount + " are available");
                        statement = statement.Substring(0, statement.Length - 2);
                        statement += whereStatement;
                        this.map.Connection.Open();
                        //Debug.Print(statement);
                        OleDbCommand updateCommand = new OleDbCommand(statement, this.map.Connection);
                        updateCommand.ExecuteNonQuery();
                        this.map.Connection.Close();
                    }
                    catch (Exception e) { // E.g. field value required, referential integrity violated.
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                            "Update row: " + e.Message, null, 0);
                        this.map.Connection.Close();
                        throw new LoggedException();
                    }
                }
            }
        }

        public void FireInsertCommand(DataRow row, Template tpl) { // "this" is the primary table.
            // Restructure with StringBuilders and field indexes as in "Retrieve auto keys" below
            if (row != null) {
                if (row.RowState != DataRowState.Added) {
                    //Debug.Print("FireInsertCommand, RowState: " + row.RowState.ToString());
                    return; // Avoid deleted and detached rows.
                }
                string statement = "INSERT INTO " + Name + "  (";
                string valueStatement = "";
                bool somethingToInsert = false;
                bool nullKey = false; // True when an autokey is generated by the database. 
                // Uvis retrieves the record and updates the key in ADO. This is necessary to set the key as a foreign key in another table.

                for (int i = 0; i < tpl.RowTpl.RowsExpr.fields.Count; i++) {
                    Field field = tpl.RowTpl.RowsExpr.fields[i];
                    if (field.prefix == tpl.RowTpl.RowsExpr.primaryTable) {
                        Column column = getColumn(field.fieldName);
                        if (row[i] is DBNull) {
                            //for (int i = 0; i < row.Table.Columns.Count; i++) {  // < tpl.Rows.fields.Count
                            // DataColumn dataColumn = row.Table.Columns[i];
                            //if (column != null && column.Prefix == tpl.Rows.primaryTable) {
                            //    if (row[dataColumn] is DBNull) {
                            if (Keys.ContainsKey(field.fieldName)) nullKey = true; // Primary key is null
                            continue; // Leave all null fields unassigned. Database may generate default values.
                        }
                        somethingToInsert = true;
                        statement += " " + column.Name + ",";
                        if (isNumber(column.ColumnType))
                            valueStatement += " " + Convert.ToString(row[i], VisEngine.codeCulture) + ", ";
                        else if (row[i] is DateTime)
                            // CodeCulture format (UK) for date is dd/MM/yyyy. This isn't accepted by the database (Access), which requires MM/dd/yyyy,
                            // but also accepts MM-dd-yyyy. The ToString below generates MM/dd/yyyy.
                            // Without codeCulture, ToString uses the local setting (e.g. - ) instead of the slashes. 
                            valueStatement += "#" + ((DateTime)row[i]).ToString("MM/dd/yy H:mm:ss", VisEngine.codeCulture) + "#, ";
                        else // String. Replace " with "".
                            valueStatement += "\"" + row[i].ToString().Replace("\"", "\"\"") + "\", ";
                    }
                }

                if (somethingToInsert) {
                    statement = statement.Substring(0, statement.Length - 1);
                    valueStatement = valueStatement.Substring(0, valueStatement.Length - 2); // Remove trailing comma_
                    statement += ")  VALUES (" + valueStatement + " )";
                    try {
                        this.map.Connection.Open();
                        //Debug.Print(statement);
                        OleDbCommand insertCommand = new OleDbCommand(statement, this.map.Connection);
                        insertCommand.ExecuteNonQuery();
                        this.map.Connection.Close();
                    }
                    catch (Exception e) { // E.g. field value required, referential integrity violated.
                        VisTool.ErrorLogging.Log(VisTool.ErrorMode.RunTime, VisTool.Priority.Error, VisEngine.curTemplate, VisEngine.curTplProp,
                            "CreateRow: " + e.Message, null, 0);
                        this.map.Connection.Close();
                        throw new LoggedException();
                    }

                    // Retrieve auto keys and defaults. SL 04-02-14. https://msdn.microsoft.com/en-us/library/ks9f57t0(v=vs.110).aspx
                    if (nullKey) { // Generate this: SELECT TOP 1 null_fields FROM primaryTable WHERE no_null_fields = values ORDER BY null_key_fields DESC;
                        // TOP 1 .  . . DESC will retrieve the largest auto keys.
                        StringBuilder select = new StringBuilder("SELECT TOP 1 ");
                        bool selectFirst = true;
                        StringBuilder where = new StringBuilder(" FROM " + tpl.RowTpl.RowsExpr.primaryTable + " WHERE ");
                        bool whereFirst = true;
                        StringBuilder orderBy = new StringBuilder(" ORDER BY ");
                        bool orderFirst = true;
                        List<int> nullFields = new List<int>(); // Indexes of fields to update.

                        for (int i = 0; i < tpl.RowTpl.RowsExpr.fields.Count; i++) {
                            Field field = tpl.RowTpl.RowsExpr.fields[i];
                            if (field.prefix == tpl.RowTpl.RowsExpr.primaryTable) {
                                Column column = getColumn(field.fieldName);
                                if (row[i] is DBNull) { // Null fields. To be retrieved if in primary table table.
                                    if (selectFirst) selectFirst = false; else select.Append(", ");
                                    select.Append(field.fieldName);
                                    nullFields.Add(i);
                                    if (Keys.ContainsKey(field.fieldName)) {
                                        if (orderFirst) orderFirst = false; else orderBy.Append(", ");
                                        orderBy.Append(column.Name + " DESC");
                                    }
                                } else { // No_null fields. Use as where conditions
                                    if (whereFirst) whereFirst = false; else where.Append(" AND ");
                                    where.Append(field.fieldName + "=");
                                    if (isNumber(column.ColumnType))
                                        where.Append(Convert.ToString(row[i], VisEngine.codeCulture));
                                    else if (row[i] is DateTime)
                                        // CodeCulture format (UK) for date is dd/MM/yyyy. This isn't accepted by the database (Access), which requires MM/dd/yyyy,
                                        // but also accepts MM-dd-yyyy. The ToString below generates MM/dd/yyyy.
                                        // Without codeCulture, ToString generates MM-dd-yyyy, using the local setting (Danish) for the dashes/slashes. 
                                        where.Append("#" + ((DateTime)row[i]).ToString("MM/dd/yy H:mm:ss", VisEngine.codeCulture) + "#");
                                    else // String. Replace single quotes with double quotes.
                                        where.Append("\"" + row[i].ToString().Replace("\"", "\"\"") + "\"");
                                }
                            } // End fields in primary table.
                        }  // End field loop

                        // Retrieve the new row from the database and insert its fields into the existing row.
                        select.Append(where.ToString()); select.Append(orderBy.ToString());
                        //Debug.Print(select.ToString());
                        OleDbDataAdapter adapter = new OleDbDataAdapter(null, VisEngine.DataMaps[0].Connection);
                        System.Data.DataSet dataset = new DataSet();
                        adapter.SelectCommand = new OleDbCommand(select.ToString(), VisEngine.DataMaps[0].Connection);
                        adapter.Fill(dataset);
                        object[] fields = dataset.Tables[0].Rows[0].ItemArray;
                        for (int i = 0; i < fields.Length; i++) if (!(fields[i] is DBNull)) row[nullFields[i]] = fields[i];
                    } // End null key
                } // End insert statement
            }
        }
    }
}
