﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;
using System.Windows.Media.Imaging;
using VisTool;
using uVisStudio; //.VisDesigner;

namespace uVisStudio {
    public static class ExtensionMethods {
        public static BitmapImage GetImage(this Bitmap image) {
            var bmp = new BitmapImage();
            var mem = new System.IO.MemoryStream();
            image.Save(mem, System.Drawing.Imaging.ImageFormat.Png);
            bmp.BeginInit();
            bmp.StreamSource = mem;
            bmp.EndInit();

            return bmp;
        }

        public static bool TryExecute(this ICommand command) {
            try {
                return command.Execute();
            }
            catch (Exception e) {
                Debug.Print("Studio Error: " + e.Message + " " + e.ToString());
                MessageBox.Show(e.Message);
                //UndoManager.PoppedCommands.Clear();
                //UndoManager.Commands.Clear();
            }
            return false;
        }

        public static bool TryUndo(this ICommand command) {
            try {
                if (!command.Undo()) return false;
                SystemItems.Refresh();
            }
            catch (Exception e) {
                Debug.Print("Studio Error: " + e.Message + " " + e.ToString());
                UndoManager.PoppedCommands.Clear();
                UndoManager.Commands.Clear();
            }
            return true;
        }

        public static Form FindForm(this VisComponent cmp) {
            VisForm frm = (VisForm)cmp.Form();
            return frm.GuiInstance as Form;
        }

        public static int GetWidth(this VisComponent instance) {
            return (instance.GuiInstance != null) ? instance.GuiInstance.ClientRectangle.Width : instance.width;
        }

        public static int GetHeight(this VisComponent instance) {
            return (instance.GuiInstance != null) ? instance.GuiInstance.ClientRectangle.Height : instance.height;
        }

        public static Template LookUp(this Template template) {
            FormTemplate frmTpl = template.FormTpl;
            if (VisEngine.Forms.ContainsKey(frmTpl.Name)) {
                Template freshTpl;
                if (frmTpl.Templates.TryGetValue(template.Name, out freshTpl))
                    return freshTpl; // The template may be a dangling part of a closed form.
            }
            MessageBox.Show(template.Name + " is closed. Cannot undo/redo", null, MessageBoxButtons.OK);
            return null;
        }
    }
}
