﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using System.Diagnostics;
using VisTool;
using uVisStudio.VisExplorer;
using System.Data;

using System.IO;

//TODO: Sometimes, when you doubleclick the boxes, their center moves to where you clicked
//that's not suposed to happen..
//TODO: The program is laggy when dragging boxes around, due to the redrawing (DrawEverything()) 
//each time the mouse moves. 

namespace uVisStudio.VisDatamap {
    /// <summary>
    /// Interaction logic for VisualDatamap.xaml
    /// </summary>
    public partial class VisualDatamapUC : UserControl {
        private List<BoxEntity> entities = new List<BoxEntity>();
        private List<Relation> relations = new List<Relation>();
        private List<RelationSelector> relationSelectors = new List<RelationSelector>();
        private Dictionary<string, BoxEntity> borderDictionary = new Dictionary<string, BoxEntity>();

        public enum resizing { No, N, NE, E, SE, S, SW, W, NW };

        private bool LayoutChanged = false;

        /// <summary>
        ///  Used primarily for the resizing of BoxEntities. 
        /// </summary>
        private class resizeStruct {
            public resizing resize_dir;
            public int curr_width, curr_height, min_width, min_height, top, bottom, left, right;
            public Point resizePoint;

            public resizeStruct() {
                resize_dir = resizing.No;
                curr_width = 0;
                curr_height = 0;
                min_width = 150;
                min_height = 20;
                top = 0;
                bottom = 0;
                left = 0;
                right = 0;
                resizePoint = new Point(-1, -1);
            }

            /// <summary>
            ///
            /// </summary>
            /// <param name="_resizing"></param> which way the resizing is to be done
            /// <param name="_curr_width"></param> the width the box had before the resizing
            /// <param name="_curr_height"></param> the height the box had before the resizing
            /// <param name="_min_width"></param> the minimum width the box can have
            /// <param name="_min_height"></param> the minimum height the box can have
            /// <param name="_top"></param> the topmost y-coord of the box before the resizing
            /// <param name="_bottom"></param> the bottommost y-coord of the box before the resizing
            /// <param name="_left"></param> the leftmost x-coord of the box before the resizing
            /// <param name="_right"></param> the rightmost x-coord of the box before the resizing
            /// <param name="_resizePoint"></param> the point where the mouse was clicked
            public resizeStruct(resizing _resize_dir, int _curr_width, int _curr_height, int _min_width,
                int _min_height, int _top, int _bottom, int _left, int _right, Point _resizePoint) {
                resize_dir = _resize_dir;
                curr_width = _curr_width;
                curr_height = _curr_height;
                min_width = _min_width;
                min_height = _min_height;
                top = _top;
                bottom = _bottom;
                left = _left;
                right = _right;
                resizePoint = _resizePoint;
            }
        }

        private resizeStruct isResizing = null;

        private Point relativeControlPosition;
        private bool isDragging = false;
        private string selectedBox;
        private int BE_MinHeight = 22;
        private int BE_MinWidth = 40;
        private Relation selectedRelation;

        //Visual fields: 
        private Color connectorColor = Colors.Black;
        private Color connectorColor_selected = Colors.Tomato;
        private Brush connectorBrush = Brushes.Black;
        private Brush boxBrush = Brushes.Ivory; //ColorConstants.WindowBackGround();
        private Brush boxBorderBrush = Brushes.DarkRed;
        private int connectorThickness = 2;
        private Thickness boxBorderThickness = new Thickness(1);
        private int circleSize = 8;

        public VisualDatamapUC() {
            InitializeComponent();
            StudioBase.datamap = this;
            this.MouseMove += CanvasMouseMove;
            this.MouseLeftButtonUp += CanvasMouseUp;
            //this.MouseRightButtonUp += RightClickOnCanvas;
            OpenFileCommand.MapFileOpened += RenameFileOpened;
            //this.SizeChanged += OnSizeChanged;
            selectedRelation = null;
            Dispatcher.ShutdownStarted += new EventHandler(Dispatcher_ShutdownStarted);
        }

        private void Dispatcher_ShutdownStarted(object sender, EventArgs e) {
                SaveSetup( );
        }

        private void RenameFileOpened(object sender, EventArgs e) {
            CreateEntities();
            DrawEverything();
            RetrieveSetup( );
        }

        private void SetCanvasSizeStandard() {
            double size = 150 * Math.Sqrt(VisEngine.DataMaps[0].getTableNames().Count);
            this.Width = size;
            this.Height = size;
            this.MinHeight = size;
            this.MinWidth = size;
        }

        public void CreateEntities() {
            this.entities.Clear();
            this.relations.Clear();
            this.borderDictionary.Clear();

            var boxEntitiesDictionary = new Dictionary<string, BoxEntity>();
            if (VisEngine.DataMaps[0] == null) return;
            var tableNames = VisEngine.DataMaps[0].getTableNames();
            if (tableNames == null) return;

            const int startLeft = 5;
            const int startTop = 5;
            var endLeft = (int)(200 * Math.Sqrt(VisEngine.DataMaps[0].getTableNames().Count)) - 150;
            var currentLeft = startLeft;
            var row = 0;
            foreach (var tableName in tableNames) {
                string key = VisTool.DataMap.PKList[tableName];

                //VisTool.Maps.Hotmap.Tables[""].getColumn

                List<string> curTFields = new List<string>();
                foreach (KeyValuePair<string, Column> c in VisEngine.DataMaps[0].Tables[tableName].Columns) {
                    if (c.Key != key)
                        curTFields.Add(c.Key);
                }

                int height_expanded = curTFields.Count * 20 + 45;
                int height_collapsed = 30;

                var boxEntity = new BoxEntity(tableName, key, currentLeft, (startTop + row * 100), 150, height_expanded, height_collapsed, curTFields);
                boxEntitiesDictionary.Add(boxEntity.Name, boxEntity);
                currentLeft += 200;

                if (currentLeft > endLeft) {
                    currentLeft = startLeft;
                    row++;
                }
            }

            foreach (var sourceName in tableNames) {
                var boxEntity = boxEntitiesDictionary[sourceName];
                foreach (var sinkName in tableNames) {
                    if (VisEngine.DataMaps[0].GetRelations(sourceName, sinkName).Count > 0) {
                        var relation = VisEngine.DataMaps[0].GetRelations(sourceName, sinkName)[0];
                        var relationBoxEntity = boxEntitiesDictionary[sinkName];
                        if (relation.Cardinality == CardinalityType.onemany) {
                            Relation rel = new Relation(relationBoxEntity, boxEntity, RelationType.onemany);
                            this.relations.Add(rel);
                        } else if (relation.Cardinality == CardinalityType.manyone) {
                            Relation rel = new Relation(relationBoxEntity, boxEntity, RelationType.manyone);
                            this.relations.Add(rel);
                        } else if (relation.Cardinality == CardinalityType.oneone) {
                            Relation rel = new Relation(relationBoxEntity, boxEntity, RelationType.oneone);
                            this.relations.Add(rel);
                        } else if (relation.Cardinality == CardinalityType.manymany) {
                            Relation rel = new Relation(relationBoxEntity, boxEntity, RelationType.manymany);
                            this.relations.Add(rel);
                        } else {
                            Relation rel = new Relation(relationBoxEntity, boxEntity, RelationType.unknown);
                            this.relations.Add(rel);
                        }
                    }
                }
                entities.Add(boxEntity);
            }

            //this next while-loop is used to ensure distinct relations
            int index = 0;
            while (index < relations.Count) {
                bool some_bool = false;
                for (int i = index + 1; i < relations.Count; i++) {
                    if (relations[index].Equals(relations[i])) some_bool = true;
                }
                if (some_bool)
                    relations.RemoveAt(index);
                else
                    index++;
            }

            foreach (Relation rel in relations) {
                RelationSelector rs = new RelationSelector(rel, new Point(0, 0), false, this);
                relationSelectors.Add(rs);
            }
        }

        public void DrawEverything() {
            //TODO: Update instead of clearing and recreating
            this.designerCanvas.Children.Clear();
            this.borderDictionary.Clear();

            //create 12 connectionpoints for each boxentity
            foreach (var boxEntity in this.entities) {
                boxEntity.MakeConnectionPoints();
            }
            //draw the connections and the icons for many-to-many, many-to-one, one-to-many and one-to-one
            foreach (var relation in this.relations) {
                DrawClosestConnectionWithConnectionPoints(relation);
            }

            //draw the boxes
            foreach (var boxEntity in this.entities) {
                if (! boxEntity.disabled) DrawBoxEntity(boxEntity);
            }
        }

        #region layout_save_retrieve

        public void RightClickOnCanvas(Object sender, MouseButtonEventArgs e) {
            if (VisEngine.ApplicationDirectory == "") return;
            System.Windows.Forms.ListView listView = new System.Windows.Forms.ListView();
            listView.Items.Add("Save layout");
            listView.Items.Add("Retrieve layout");
            listView.Items.Add("Cancel");
            listView.MultiSelect = false;
            listView.Height = 48;
            listView.View = System.Windows.Forms.View.SmallIcon;
            listView.Activation = System.Windows.Forms.ItemActivation.OneClick;

            listView.ItemActivate += new EventHandler(setup_listView_ItemActivate);

            System.Windows.Forms.ToolStripDropDown popup = new System.Windows.Forms.ToolStripDropDown();
            popup.ImageScalingSize = new System.Drawing.Size(0, 0);
            popup.ShowItemToolTips = true;
            popup.AutoSize = false;
            popup.Margin = System.Windows.Forms.Padding.Empty;
            popup.Padding = System.Windows.Forms.Padding.Empty;
            System.Windows.Forms.ToolStripControlHost host = new System.Windows.Forms.ToolStripControlHost(listView);
            host.Margin = System.Windows.Forms.Padding.Empty;
            host.Padding = System.Windows.Forms.Padding.Empty;
            host.AutoSize = false;
            host.Size = listView.Size;
            host.ToolTipText = "\"Save setup\" saves the current positions and sizes of \n" +
                                "the tables (the boxes representing database tables) for \n " +
                                "recreating of this setup in another session of Vistool.\n" +
                                "\n\"Retrieve setup\" loads the setup that was previously saved \n" +
                                "for this datamap, and repositions and resizes the boxes.";
            popup.Size = listView.Size;
            popup.Items.Add(host);

            //find the point where we want the popup to show
            var pos = this.PointToScreen(Mouse.GetPosition(this));

            //show the popup
            popup.Show((int)pos.X, (int)pos.Y);
        }

        public void setup_listView_ItemActivate(Object sender, EventArgs e) {
            var listview = sender as System.Windows.Forms.ListView;
            var item = listview.SelectedItems[0].ToString();
            item = item.Substring(15, item.Length - 16);
            if (item.Equals("Save setup")) {
                this.SaveSetup( );
            } else if (item.Equals("Retrieve setup")) {
                this.RetrieveSetup( );
            }
            //find the popup and kill it!
            var popup = listview.Parent as System.Windows.Forms.ToolStripDropDown;
            popup.Close();
        }

        public void Clean() { // Remove all entities and relations, e.g. to load a new datamap.
            SaveSetup( );
            relations.Clear();
            entities.Clear();
        }

        /// <summary>
        /// Save or retrieve the current datamap layout
        /// </summary>
        /// <param name="result"> Pass 1 to load the last saved setup. Pass 2 to save the current setup. </param>
        public void SaveSetup() {
            if (VisEngine.ApplicationDirectory == "" || !LayoutChanged) return;
            MessageBoxResult res = MessageBox.Show("Save the datamap layout?", "Save datamap layout?", MessageBoxButton.YesNo);
            if (res == MessageBoxResult.No) return;

            StreamWriter tw = new StreamWriter(VisEngine.ApplicationDirectory + "\\" + VisEngine.MapFile + ".visER");
            tw.WriteLine("' Positioning and sizing of box entities for \"" + VisEngine.MapFile + ".vis\"\n' Created " + DateTime.Now);
            tw.Close();

            using (StreamWriter w = File.AppendText(VisEngine.ApplicationDirectory + "\\" + VisEngine.MapFile + ".visER")) {
                w.WriteLine("-------------------------------------------------");
                foreach (BoxEntity box in entities) {
                    writeBoxEntity(box, w);
                }
                w.Close();
            }
        }

        public void RetrieveSetup() {
            LayoutChanged = false;
            try {
                using (StreamReader r = new StreamReader(VisEngine.ApplicationDirectory + "\\" + VisEngine.MapFile + ".visER")) {
                    ReadvisER(r);
                }
            }
            catch (Exception) { // File may not exist. Don't care.
            }
            DrawEverything();
        }

        private void writeBoxEntity(BoxEntity box, StreamWriter stream) {
            stream.WriteLine("BoxEntity: " + box.Name);
            stream.WriteLine("Top: " + box.Top);
            stream.WriteLine("Left: " + box.Left);
            stream.WriteLine("Width: " + box.Width);
            stream.WriteLine("Height: " + box.Height);
            stream.WriteLine("fieldsVisible: " + box.FieldsVisible);
            stream.WriteLine("height_expanded: " + box.height_expanded);
            stream.WriteLine("height_collapsed: " + box.height_collapsed);
            stream.WriteLine("-------------------------------------------------");
            //Update the file
            stream.Flush();
        }

        private void ReadvisER(StreamReader stream) {
            string line;
            BoxEntity box = null;
            List<BoxEntity> boxes = entities.ToList();
            while ((line = stream.ReadLine()) != null) {
                if (line.StartsWith("-") || line.StartsWith("'")) //these lines are uninsteresting for this purpose
                    continue;
                string[] arr = line.Split(new string[] { ": " }, StringSplitOptions.RemoveEmptyEntries);
                //switch on the first element of the line - being a property or a name
                try {
                    switch (arr[0]) {
                    case "BoxEntity":
                        box = boxes.Find(i => i.Name == arr[1]); //find the boxentity that the line specifies
                        break;
                    case "Top":
                        if (box != null)
                            box.Top = Int32.Parse(arr[1]);
                        break;
                    case "Left":
                        if (box != null)
                            box.Left = Int32.Parse(arr[1]);
                        break;
                    case "Width":
                        if (box != null)
                            box.Width = Int32.Parse(arr[1]);
                        break;
                    case "Height":
                        if (box != null)
                            box.Height = Int32.Parse(arr[1]);
                        break;
                    case "fieldsVisible":
                        if (box != null)
                            box.FieldsVisible = Boolean.Parse(arr[1]);
                        break;
                    case "height_expanded":
                        if (box != null)
                            box.height_expanded = Int32.Parse(arr[1]);
                        break;
                    case "height_collapsed":
                        if (box != null)
                            box.height_collapsed = Int32.Parse(arr[1]);
                        break;
                    default:

                        break;
                    }
                }
                catch (Exception e) {
                    //TODO: Do some error logging here instead
                    System.Console.WriteLine("Something went wrong in ReadvisER() " + e.ToString());
                }
            }
            stream.Close();
        }

        #endregion layout_save_retrieve

        #region EVENTHANDLERS

        public void ExpandButtonClicked(Object sender, MouseButtonEventArgs e) {
            LayoutChanged = true;
            Border border = (Border)((Panel)((StackPanel)((Image)sender).Parent).Parent).Parent;
            string name = border.Name;
            BoxEntity theBoxE = (BoxEntity)border.Tag;
            //TODO: Something like error.log here instead
            if (theBoxE == null) { MessageBox.Show("There is no BoxEntity with the name: " + name); return; }

            theBoxE.FieldsVisible = !theBoxE.FieldsVisible;
            theBoxE.last_clicked = DateTime.MinValue;

            isDragging = false;
            isResizing = null;
            DrawEverything();

            e.Handled = true;
        }

        public void MouseEnterExpandButtonBoxEntity(Object sender, EventArgs e) {
            if (Mouse.LeftButton != MouseButtonState.Pressed)
                this.Cursor = Cursors.Hand;
        }

        public void MouseExitExpandButtonBoxEntity(Object sender, EventArgs e) {
            if (Mouse.LeftButton != MouseButtonState.Pressed)
                this.Cursor = Cursors.Arrow;
        }

        private void CanvasMouseMove(object sender, MouseEventArgs e) {
            if (Mouse.LeftButton == MouseButtonState.Pressed) {
                if (this.selectedBox == null) return; //security check..
                BoxEntity box = this.borderDictionary[this.selectedBox];

                //TODO: (this.isResizing == null || this.isResizing.resize_dir == resizing.No) should be unnecessary
                if ((this.isResizing == null || this.isResizing.resize_dir == resizing.No) && isDragging) {
                    Point newMousePosition = Mouse.GetPosition(designerCanvas);
                    if (box.Left != (int)newMousePosition.X - (int)this.relativeControlPosition.X) {
                        box.Left = (int)newMousePosition.X - (int)this.relativeControlPosition.X;
                        LayoutChanged = true;
                    }
                    if (box.Top != (int)newMousePosition.Y - (int)this.relativeControlPosition.Y) {
                        box.Top = (int)newMousePosition.Y - (int)this.relativeControlPosition.Y;
                        LayoutChanged = true;
                    }
                    DrawEverything();

                } else if (this.isResizing != null && this.isResizing.resize_dir != resizing.No) //we are resizing a BoxEntity
                {
                    Point newMousePosition = Mouse.GetPosition(this);
                    if (this.isResizing.resize_dir == resizing.N || this.isResizing.resize_dir == resizing.NE || this.isResizing.resize_dir == resizing.NW) {
                        int newheight = (int)this.isResizing.resizePoint.Y - (int)newMousePosition.Y + this.isResizing.curr_height;
                        if (!(newheight < BE_MinHeight) && box.Height != newheight) { box.Height = newheight; box.Top = (int)newMousePosition.Y; LayoutChanged = true; }
                    }
                    if (this.isResizing.resize_dir == resizing.S || this.isResizing.resize_dir == resizing.SE || this.isResizing.resize_dir == resizing.SW) {
                        int newheight = (int)newMousePosition.Y - (int)this.isResizing.resizePoint.Y + this.isResizing.curr_height;
                        newheight = newheight < BE_MinHeight ? BE_MinHeight : newheight;
                        if (box.Height != newheight) { box.Height = newheight; LayoutChanged = true; }
                    }
                    if (this.isResizing.resize_dir == resizing.W || this.isResizing.resize_dir == resizing.SW || this.isResizing.resize_dir == resizing.NW) {
                        int newwidth = (int)this.isResizing.resizePoint.X - (int)newMousePosition.X + this.isResizing.curr_width;
                        if (!(newwidth < BE_MinWidth) && box.Width != newwidth) { box.Width = newwidth; box.Left = (int)newMousePosition.X; LayoutChanged = true; }
                    }
                    if (this.isResizing.resize_dir == resizing.E || this.isResizing.resize_dir == resizing.SE || this.isResizing.resize_dir == resizing.NE) {
                        int newwidth = (int)newMousePosition.X - (int)this.isResizing.resizePoint.X + this.isResizing.curr_width;
                        newwidth = newwidth < BE_MinWidth ? BE_MinWidth : newwidth;
                        if (box.Width != newwidth) { box.Width = newwidth; LayoutChanged = true; }
                    }
                    if (box.FieldsVisible)
                        box.height_expanded = box.Height;
                    else
                        box.height_collapsed = box.Height;

                    DrawEverything();
                }
            } else {
                this.selectedBox = null;
            }
        }

        public void CanvasMouseUp(object sender, MouseEventArgs e) {
            DrawEverything();
            isResizing = null;
            isDragging = false;
        }

        private void BorderMouseHoverExit(object sender, MouseEventArgs e) {
            if (Mouse.LeftButton != MouseButtonState.Pressed)
                this.Cursor = Cursors.Arrow;
        }

        /// <summary>
        /// Only used to change the image of the mouse when hovering an edge of a BoxEntity
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BorderMouseHoverMargin(object sender, MouseEventArgs e) {
            if (this.isResizing != null || this.isDragging) return;
            Border border = (Border)sender;
            BoxEntity box = (BoxEntity)border.Tag;
            Point mousePos = Mouse.GetPosition(this);
            resizeStruct res = new resizeStruct();
            res.top = box.Top;
            res.left = box.Left;
            res.right = box.Left + box.Width;
            res.bottom = box.Top + box.Height;

            int margin = 6;
            if (Math.Abs(mousePos.Y - res.top) < margin) //top edge
                res.resize_dir = resizing.N;
            if (Math.Abs(mousePos.Y - res.bottom) < margin) //bottom edge
                res.resize_dir = resizing.S;
            if (Math.Abs(mousePos.X - res.left) < margin) //left edge
                res.resize_dir = res.resize_dir == resizing.N ? resizing.NW : (res.resize_dir == resizing.S ? resizing.SW : resizing.W);
            if (Math.Abs(mousePos.X - res.right) < margin) //right edge
                res.resize_dir = res.resize_dir == resizing.N ? resizing.NE : (res.resize_dir == resizing.S ? resizing.SE : resizing.E);
            if (res.resize_dir == resizing.N || res.resize_dir == resizing.S)
                this.Cursor = Cursors.SizeNS;
            else if (res.resize_dir == resizing.W || res.resize_dir == resizing.E)
                this.Cursor = Cursors.SizeWE;
            else if (res.resize_dir == resizing.NW || res.resize_dir == resizing.SE)
                this.Cursor = Cursors.SizeNWSE;
            else if (res.resize_dir == resizing.NE || res.resize_dir == resizing.SW)
                this.Cursor = Cursors.SizeNESW;
            else if (Mouse.LeftButton != MouseButtonState.Pressed && this.Cursor != Cursors.Hand)
                //TODO: This aint that brilliant. Something better would be to check on the movement
                //of the mouse itself, and then identifying which control is underneath it
                this.Cursor = Cursors.Arrow;
        }

        private void BorderMouseLeftButtonDown(object sender, MouseButtonEventArgs e) {
            Border border = (Border)sender;
            BoxEntity box = (BoxEntity)border.Tag;
            entities.Remove(box);
            entities.Add(box);
            Point mousePos = Mouse.GetPosition(this);
            this.selectedBox = border.Name;

            resizeStruct res = new resizeStruct();
            res.resizePoint = mousePos;
            res.curr_height = box.Height;
            res.curr_width = box.Width;
            //magic numbers, we luves it: their origin is in CreateEntities()
            res.min_height = 30;
            res.min_width = 150;
            res.top = box.Top;
            res.left = box.Left;
            res.right = box.Left + box.Width;
            res.bottom = box.Top + box.Height;

            int margin = 6;
            if (Math.Abs(mousePos.Y - res.top) < margin) //top edge
                res.resize_dir = resizing.N;
            if (Math.Abs(mousePos.Y - res.bottom) < margin) //bottom edge
                res.resize_dir = resizing.S;
            if (Math.Abs(mousePos.X - res.left) < margin) //left edge
                res.resize_dir = res.resize_dir == resizing.N ? resizing.NW : (res.resize_dir == resizing.S ? resizing.SW : resizing.W);
            if (Math.Abs(mousePos.X - res.right) < margin) //right edge
                res.resize_dir = res.resize_dir == resizing.N ? resizing.NE : (res.resize_dir == resizing.S ? resizing.SE : resizing.E);

            if (res.resize_dir != resizing.No)
                this.isResizing = res;
            else {
                //TODO: one of these might not be necessary
                // they were made before I knew how to stop the bubbling of events
                this.isResizing = null;
                isDragging = true;
            }

            if (res.resize_dir == resizing.N || res.resize_dir == resizing.S)
                Cursor = Cursors.SizeNS;
            else if (res.resize_dir == resizing.W || res.resize_dir == resizing.E)
                Cursor = Cursors.SizeWE;
            else if (res.resize_dir == resizing.NW || res.resize_dir == resizing.SE)
                Cursor = Cursors.SizeNWSE;
            else if (res.resize_dir == resizing.NE || res.resize_dir == resizing.SW)
                Cursor = Cursors.SizeNESW;
            else //isResizing == resizing.No
            {
                this.relativeControlPosition = Mouse.GetPosition(border);
            }

            ((BoxEntity)((Border)sender).Tag).MouseDoubleClick(DateTime.Now);
        }

        private void BorderMouseButtonUp(object sender, MouseButtonEventArgs e) {
            if (this.selectedBox == null) return; //security check..
            isDragging = false;
            this.Cursor = Cursors.Arrow;

            this.Cursor = Cursors.Arrow;
            BoxEntity box = this.borderDictionary[selectedBox];

            if (box.FieldsVisible)
                box.height_expanded = box.Height;
            else
                box.height_collapsed = box.Height;
            DrawEverything();

        }

        #endregion EVENTHANDLERS

        public void DrawBoxEntity(BoxEntity boxEntity) {
            try {
                var border = new Border {
                    Name = boxEntity.Name,
                    Width = boxEntity.Width,
                    Height = boxEntity.Height,
                    Background = boxBrush,
                    CornerRadius = new CornerRadius(0),
                    BorderBrush = boxBorderBrush,
                    BorderThickness = boxBorderThickness,
                    Tag = boxEntity
                };
                var panel = new StackPanel { VerticalAlignment = VerticalAlignment.Stretch, Margin = new Thickness(0), Orientation = System.Windows.Controls.Orientation.Vertical };
                var namePanel = new StackPanel { VerticalAlignment = VerticalAlignment.Stretch, HorizontalAlignment = HorizontalAlignment.Left, Margin = new Thickness(0), Orientation = System.Windows.Controls.Orientation.Horizontal };
                var label = new System.Windows.Controls.Label() {
                    Content = boxEntity.Name.Replace("_", "__"),
                    VerticalContentAlignment = VerticalAlignment.Center,
                    HorizontalContentAlignment = HorizontalAlignment.Center,
                    FontWeight = FontWeights.Bold,
                    Margin = new Thickness(0),
                    Padding = new Thickness(0)
                };
                var keyPanel = new StackPanel { Orientation = System.Windows.Controls.Orientation.Vertical, Margin = new Thickness(0) };
                ImageSource i_s = boxEntity.FieldsVisible ? Properties.Resources.treeview_collapse.GetImage() : Properties.Resources.treeview_expand.GetImage();
                //The below is somehow causing the canvas to be laggy, really laggy
                //ImageSource i_s = boxEntity.FieldsVisible ? Properties.Resources.Minus_Green_Button.GetImage() : Properties.Resources.Add_Green_Button.GetImage();
                var i = new Image { Stretch = Stretch.None, Width = 16, Height = 16, Margin = new Thickness(2), Source = i_s };

                i.MouseLeftButtonDown += new MouseButtonEventHandler(ExpandButtonClicked);
                i.MouseEnter += new MouseEventHandler(MouseEnterExpandButtonBoxEntity);
                i.MouseLeave += new MouseEventHandler(MouseExitExpandButtonBoxEntity);

                border.MouseLeftButtonDown += new MouseButtonEventHandler(BorderMouseLeftButtonDown);
                border.MouseUp += new MouseButtonEventHandler(BorderMouseButtonUp);
                border.MouseMove += new MouseEventHandler(BorderMouseHoverMargin);
                border.MouseLeave += new MouseEventHandler(BorderMouseHoverExit);

                //TODO: Make it a little transparent in order to see connectors going through it
                border.Opacity = 0.9;

                border.Child = panel;
                namePanel.Children.Add(i);
                namePanel.Children.Add(label);
                panel.Children.Add(namePanel);
                panel.Children.Add(keyPanel);

                this.borderDictionary.Add(border.Name, boxEntity);
                if (boxEntity.fieldsVisible == true) {
                    System.Windows.Controls.Label pk = new System.Windows.Controls.Label() {
                        Padding = new Thickness(0), Margin = new Thickness(2),
                        Content = boxEntity.PrimaryKey.Replace("_", "__")
                    };
                    keyPanel.Children.Add(pk);
                    foreach (string s in boxEntity.Fields) {
                        System.Windows.Controls.Label lblField = new System.Windows.Controls.Label() {
                            Padding = new Thickness(0), Margin = new Thickness(2),
                            Content = s.Replace("_", "__")
                        }; // A single underscore becomes a short-cut key and doesn't show.
                        keyPanel.Children.Add(lblField);
                    }
                }
                System.Windows.Controls.Canvas.SetLeft(border, boxEntity.Left);
                System.Windows.Controls.Canvas.SetTop(border, boxEntity.Top);
                designerCanvas.Children.Add(border);
            }
            catch (Exception) {
                boxEntity.disabled = true;
                Debug.Print("Could not draw box corresponding to table: " + boxEntity.Name);
            }
        }

        private PathGeometry DrawLine(System.Windows.Controls.Canvas canvas, double X1, double Y1, double X2, double Y2, PointPosition fromPosition, PointPosition toPosition, Color color) {
            var xDiff = Math.Abs(X2 - X1);
            var yDiff = Math.Abs(Y2 - Y1);
            var tangent = Math.Min(xDiff, yDiff);

            var sourceControlPoint = new Point();
            var sinkControlPoint = new Point();
            if (fromPosition == PointPosition.Left) sourceControlPoint = new Point(X1 - tangent, Y1);
            else if (fromPosition == PointPosition.Right) sourceControlPoint = new Point(X1 + tangent, Y1);
            else if (fromPosition == PointPosition.Top) sourceControlPoint = new Point(X1, Y1 - tangent);
            else if (fromPosition == PointPosition.Bottom) sourceControlPoint = new Point(X1, Y1 + tangent);

            if (toPosition == PointPosition.Left) sinkControlPoint = new Point(X2 - tangent, Y2);
            else if (toPosition == PointPosition.Right) sinkControlPoint = new Point(X2 + tangent, Y2);
            else if (toPosition == PointPosition.Top) sinkControlPoint = new Point(X2, Y2 - tangent);
            else if (toPosition == PointPosition.Bottom) sinkControlPoint = new Point(X2, Y2 + tangent);

            var bezierSegment = new BezierSegment(sourceControlPoint, sinkControlPoint, new Point(X2, Y2), true);

            var pscollection = new PathSegmentCollection();
            pscollection.Add(bezierSegment);
            var pf1 = new PathFigure { Segments = pscollection, StartPoint = new Point(X1, Y1) };
            var pfcollection = new PathFigureCollection();
            pfcollection.Add(pf1);
            var pathGeometry = new PathGeometry { Figures = pfcollection };

            var path = new System.Windows.Shapes.Path { Data = pathGeometry, Stroke = new SolidColorBrush(color), StrokeThickness = 2 };
            canvas.Children.Add(path);

            return pathGeometry;
        }

        private void DrawClosestConnectionWithConnectionPoints(Relation relation) {
            var distance = double.MaxValue;
            double minDistance = 15.0;
            var source = relation.Source;
            var sink = relation.Sink;
            var sourceEntityPoint = source.ConnectionPoints[0];
            var sinkEntityPoint = sink.ConnectionPoints[0];
            foreach (var boxEntityPoint1 in source.ConnectionPoints) {
                var point1 = boxEntityPoint1.Point;
                foreach (var boxEntityPoint2 in sink.ConnectionPoints) {
                    var point2 = boxEntityPoint2.Point;
                    var newDistance = Math.Sqrt(Math.Pow(point2.X - point1.X, 2) + Math.Pow(point2.Y - point1.Y, 2));
                    if (newDistance < distance && newDistance > minDistance) {
                        distance = newDistance;
                        sourceEntityPoint = boxEntityPoint1;
                        sinkEntityPoint = boxEntityPoint2;
                    }
                }
            }

            Point sourcePoint = sourceEntityPoint.Point;
            Point sinkPoint = sinkEntityPoint.Point;
            double sinkX = sinkPoint.X, sinkY = sinkPoint.Y, sourceX = sourcePoint.X, sourceY = sourcePoint.Y;

            int correction = circleSize / 2;
            if (relation.RelationType == RelationType.manyone || relation.RelationType == RelationType.manymany)
                correction = 10;
            else
                correction = circleSize / 2;

            if (sourceEntityPoint.PointPosition == PointPosition.Left)
                sourceX -= correction;
            if (sourceEntityPoint.PointPosition == PointPosition.Top)
                sourceY -= correction;
            if (sourceEntityPoint.PointPosition == PointPosition.Right)
                sourceX += correction;
            if (sourceEntityPoint.PointPosition == PointPosition.Bottom)
                sourceY += correction;

            if (relation.RelationType == RelationType.onemany || relation.RelationType == RelationType.manymany)
                correction = 10;
            else
                correction = circleSize / 2;

            if (sinkEntityPoint.PointPosition == PointPosition.Left)
                sinkX -= correction;
            if (sinkEntityPoint.PointPosition == PointPosition.Top)
                sinkY -= correction;
            if (sinkEntityPoint.PointPosition == PointPosition.Right)
                sinkX += correction;
            if (sinkEntityPoint.PointPosition == PointPosition.Bottom)
                sinkY += correction;

            //TODO: MoveRelationSelector?
            if (selectedRelation != null && selectedRelation.Sink == relation.Sink && selectedRelation.Source == relation.Source)
                this.DrawLine(designerCanvas, sourceX, sourceY, sinkX, sinkY, sourceEntityPoint.PointPosition, sinkEntityPoint.PointPosition, connectorColor_selected);
            else
                this.DrawLine(designerCanvas, sourceX, sourceY, sinkX, sinkY, sourceEntityPoint.PointPosition, sinkEntityPoint.PointPosition, connectorColor);

            DrawConnectionPoints(relation);

            source.ConnectionPoints.Remove(sourceEntityPoint);
            sink.ConnectionPoints.Remove(sinkEntityPoint);
        }

        public class RelationSelector {
            Relation relation;
            Point position;
            bool isSource;
            VisualDatamapUC datamap;

            /// <summary>
            /// </summary>
            /// <param name="rel"> the relation to which this relationselector belongs </param>
            /// <param name="position_center"> the position of this relationselector</param>
            /// <param name="isSource"> whether this relationselector is the source of the relation (if false, it is assumed to be the sink)</param>
            public RelationSelector(Relation rel, Point position_center, bool isSource, VisualDatamapUC datamap) {
                this.relation = rel;
                this.position = position_center;
                this.isSource = isSource;
                this.datamap = datamap;
            }

            public void ClickMe(Object sender, MouseButtonEventArgs e) {
                datamap.selectedRelation = this.relation;
            }

            public void Draw() {
                //create a border
                Border border = new Border {
                    Name = this.ToString() + isSource.ToString().ToLower(),
                    Width = 10,
                    Height = 10,
                    Background = Brushes.White,
                    CornerRadius = new CornerRadius(0),
                    BorderBrush = Brushes.Tomato,
                    BorderThickness = new Thickness(1),
                    Tag = this
                };


                //add an eventhandler 
                border.MouseLeftButtonDown += ClickMe;

                //add the border to the designercanvas
                datamap.designerCanvas.Children.Add(border);
                System.Windows.Controls.Canvas.SetLeft(border, position.X - 5);
                System.Windows.Controls.Canvas.SetTop(border, position.Y - 5);
            }
        }

        private void DrawConnectionPoints(Relation relation) {
            var distance = double.MaxValue;
            var source = relation.Source;
            var sink = relation.Sink;
            var sourceEntityPoint = source.ConnectionPoints[0];
            var sinkEntityPoint = sink.ConnectionPoints[0];
            foreach (var boxEntityPoint1 in source.ConnectionPoints) {
                var point1 = boxEntityPoint1.Point;
                foreach (var boxEntityPoint2 in sink.ConnectionPoints) {
                    var point2 = boxEntityPoint2.Point;
                    var newDistance = Math.Sqrt(Math.Pow(point2.X - point1.X, 2) + Math.Pow(point2.Y - point1.Y, 2));
                    if (newDistance < distance) {
                        distance = newDistance;
                        sourceEntityPoint = boxEntityPoint1;
                        sinkEntityPoint = boxEntityPoint2;
                    }
                }
            }
            bool many = false;
            if (relation.RelationType == RelationType.manymany || relation.RelationType == RelationType.manyone)
                many = true;
            this.DrawPoint(sourceEntityPoint, many);
            if (relation.RelationType == RelationType.onemany || relation.RelationType == RelationType.manymany)
                many = true;
            else
                many = false;
            this.DrawPoint(sinkEntityPoint, many);

            //if(relation.RelationType == RelationType.manymany || relation.RelationType == RelationType.manyone)
            //    this.DrawPoint(sourceEntityPoint, RelationType.manyone);
            //else
            //    this.DrawPoint(sourceEntityPoint, RelationType.onemany);

            //if (relation.RelationType == RelationType.manymany || relation.RelationType == RelationType.onemany)
            //    this.DrawPoint(sinkEntityPoint, RelationType.onemany);
            //else
            //    this.DrawPoint(sinkEntityPoint, RelationType.manyone);

            source.ConnectionPoints.Remove(sourceEntityPoint);
            sink.ConnectionPoints.Remove(sinkEntityPoint);
        }

        private void DrawPoint(BoxEntityPoint sourceEntityPoint, bool many) {
            Point sourcePoint = sourceEntityPoint.Point;
            int rotationDegree = 0;
            int xMove = 0;
            int yMove = 0;
            if (sourceEntityPoint.PointPosition == PointPosition.Top) {
                rotationDegree = 0;
                yMove = -10;
            } else if (sourceEntityPoint.PointPosition == PointPosition.Right) {
                rotationDegree = 90;
                xMove = 10;
            } else if (sourceEntityPoint.PointPosition == PointPosition.Bottom) {
                rotationDegree = 180;
                yMove = 10;
            } else if (sourceEntityPoint.PointPosition == PointPosition.Left) {
                rotationDegree = 270;
                xMove = -10;
            }

            if (many) {
                var l = new Line();
                l.X1 = sourcePoint.X;
                l.Y1 = sourcePoint.Y;
                l.X2 = sourcePoint.X;
                l.Y2 = sourcePoint.Y + 10;
                l.Stroke = connectorBrush;
                l.StrokeThickness = connectorThickness;

                var lLeft = new Line();
                lLeft.X1 = sourcePoint.X;
                lLeft.Y1 = sourcePoint.Y;
                lLeft.X2 = sourcePoint.X - 5;
                lLeft.Y2 = sourcePoint.Y + 10;
                lLeft.Stroke = connectorBrush;
                lLeft.StrokeThickness = connectorThickness;

                var lRight = new Line();
                lRight.X1 = sourcePoint.X;
                lRight.Y1 = sourcePoint.Y;
                lRight.X2 = sourcePoint.X + 5;
                lRight.Y2 = sourcePoint.Y + 10;
                lRight.Stroke = connectorBrush;
                lRight.Stroke = connectorBrush;
                lRight.StrokeThickness = connectorThickness;

                var translate = new TranslateTransform(xMove, yMove);
                var rotation = new RotateTransform(rotationDegree, lLeft.X1, lLeft.Y1);
                var transformGroup = new TransformGroup();
                transformGroup.Children.Add(rotation);
                transformGroup.Children.Add(translate);

                lLeft.RenderTransform = transformGroup;
                l.RenderTransform = transformGroup;
                lRight.RenderTransform = transformGroup;

                designerCanvas.Children.Add(l);
                designerCanvas.Children.Add(lLeft);
                designerCanvas.Children.Add(lRight);
            } else if (!many) {
                System.Windows.Shapes.Ellipse circle = new System.Windows.Shapes.Ellipse();
                circle.StrokeThickness = connectorThickness;
                circle.Stroke = connectorBrush;
                circle.Fill = connectorBrush;
                circle.HorizontalAlignment = HorizontalAlignment.Center;
                circle.VerticalAlignment = VerticalAlignment.Center;
                circle.Width = circleSize;
                circle.Height = circleSize;
                positionEllipse(circle, circle.Width, circle.Height,
                    sourcePoint.X + xMove / 2 + (xMove < 0 ? circleSize / 2 : xMove > 0 ? -circleSize / 2 : 0),
                    sourcePoint.Y + yMove / 2 + (yMove < 0 ? circleSize / 2 : yMove > 0 ? -circleSize / 2 : 0));
                designerCanvas.Children.Add(circle);
            }
        }

        private void positionEllipse(System.Windows.Shapes.Ellipse ellipse, double width, double height, double desiredCenterX, double desiredCenterY) {
            double left = desiredCenterX - (width / 2);
            double top = desiredCenterY - (height / 2);
            ellipse.Margin = new Thickness(left, top, 0, 0);
        }
    }

    //TODO: Move the classes below to their own files. 900 lines of code is already enough in one class.
    public class DesignerCanvas : System.Windows.Controls.Canvas {
        protected override Size MeasureOverride(Size constraint) {
            var size = new Size();
            foreach (UIElement element in Children) {
                double left = GetLeft(element);
                double top = GetTop(element);
                left = double.IsNaN(left) ? 0 : left;
                top = double.IsNaN(top) ? 0 : top;

                element.Measure(constraint);

                Size desiredSize = element.DesiredSize;
                if (!double.IsNaN(desiredSize.Width) && !double.IsNaN(desiredSize.Height)) {
                    size.Width = Math.Max(size.Width, left + desiredSize.Width);
                    size.Height = Math.Max(size.Height, top + desiredSize.Height);
                }
            }

            // add some extra margin
            size.Width += 10;
            size.Height += 10;
            return size;
        }
    }

    public class BoxEntity {
        public bool disabled = false; // True if the box entity shouldn't be shown.
        private int left;
        private int top;
        private int width;
        private int height;
        //Lower bound for the speed at which a double-click can be performed
        private double double_click_speed = 0.7f;
        public bool fieldsVisible { get; set; }
        //This field is used for acknowledging double-clicks. See MouseDoubleClick()
        public DateTime last_clicked { get; set; }

        #region properties

        public string Name { get; set; }
        public string PrimaryKey { get; set; }
        public List<BoxEntityPoint> ConnectionPoints { get; set; }
        public List<string> Fields { get; set; }
        public int height_expanded { get; set; }
        public int height_collapsed { get; set; }

        public bool FieldsVisible {
            get { return this.fieldsVisible; }
            set { this.fieldsVisible = value; height = fieldsVisible ? height_expanded : height_collapsed; }
        }

        public int Left {
            get { return this.left; }
            set { this.left = value; this.MakeConnectionPoints(); }
        }

        public int Top {
            get { return this.top; }
            set { this.top = value; this.MakeConnectionPoints(); }
        }

        public int Width {
            get { return this.width; }
            set { this.width = value; this.MakeConnectionPoints(); }
        }

        public int Height {
            get { return this.height; }
            set { this.height = value; this.MakeConnectionPoints(); }
        }

        #endregion

        public BoxEntity(string name, string primaryKey, int left, int top, int width, int height_expanded, int height_collapsed, List<string> Fields) {
            ConnectionPoints = new List<BoxEntityPoint>();
            this.Name = name;
            this.PrimaryKey = primaryKey;
            this.left = left;
            this.top = top;
            this.width = width;
            this.Fields = Fields;
            this.MakeConnectionPoints();

            this.height_expanded = height_expanded;
            this.height_collapsed = height_collapsed;
            this.FieldsVisible = false;
        }

        /// <summary>
        /// Called when the boxentity is clicked. Used to determine whether the boxentity
        /// was double clicked.
        /// </summary>
        /// <param name="click_time"></param> When was the border holding this BoxEntity clicked?
        public void MouseDoubleClick(DateTime time_of_the_click) {
            //Was the boxEntity was double-clicked?
            if (last_clicked > time_of_the_click.AddSeconds(-double_click_speed)) {
                last_clicked = time_of_the_click.AddSeconds(-5);
                DataViewForm dvtf2 = new DataViewForm();
                dvtf2.Show();
                dvtf2.FillData(Name);
                dvtf2.Activate();
            } else {
                last_clicked = time_of_the_click;
            }
        }

        public Point GetCenter() {
            var centerX = this.left + (this.width / 2);
            var centerY = this.top + (this.height / 2);
            return new Point(centerX, centerY);
        }

        public void MakeConnectionPoints() {
            var leftPoint = new BoxEntityPoint(new Point(this.left, this.top + (this.height / 2)), PointPosition.Left);
            var leftTopPoint = new BoxEntityPoint(new Point(this.left, this.top + (this.height / 4)), PointPosition.Left);
            var leftBottomPoint = new BoxEntityPoint(new Point(this.left, this.top + (3 * this.height / 4)), PointPosition.Left);
            var topPoint = new BoxEntityPoint(new Point(this.left + (this.width / 2), this.top), PointPosition.Top);
            var topLeftPoint = new BoxEntityPoint(new Point(this.left + (this.width / 4), this.top), PointPosition.Top);
            var topRightPoint = new BoxEntityPoint(new Point(this.left + (3 * this.width / 4), this.top), PointPosition.Top);
            var rightPoint = new BoxEntityPoint(new Point(this.left + this.width, this.top + (this.height / 2)), PointPosition.Right);
            var rightTopPoint = new BoxEntityPoint(new Point(this.left + this.width, this.top + (this.height / 4)), PointPosition.Right);
            var rightBottomPoint = new BoxEntityPoint(new Point(this.left + this.width, this.top + (3 * this.height / 4)), PointPosition.Right);
            var bottomPoint = new BoxEntityPoint(new Point(this.left + (this.width / 2), this.top + this.height), PointPosition.Bottom);
            var bottomLeftPoint = new BoxEntityPoint(new Point(this.left + (this.width / 4), this.top + this.height), PointPosition.Bottom);
            var bottomRightPoint = new BoxEntityPoint(new Point(this.left + (3 * this.width / 4), this.top + this.height), PointPosition.Bottom);

            this.ConnectionPoints = new List<BoxEntityPoint>();
            this.ConnectionPoints.Add(leftPoint);
            this.ConnectionPoints.Add(leftTopPoint);
            this.ConnectionPoints.Add(leftBottomPoint);
            this.ConnectionPoints.Add(topPoint);
            this.ConnectionPoints.Add(topLeftPoint);
            this.ConnectionPoints.Add(topRightPoint);
            this.ConnectionPoints.Add(rightPoint);
            this.ConnectionPoints.Add(rightTopPoint);
            this.ConnectionPoints.Add(rightBottomPoint);
            this.ConnectionPoints.Add(bottomPoint);
            this.ConnectionPoints.Add(bottomLeftPoint);
            this.ConnectionPoints.Add(bottomRightPoint);
        }
    }

    public enum PointPosition {
        Left,
        Top,
        Right,
        Bottom
    }

    public class BoxEntityPoint {
        public Point Point { get; set; }
        public PointPosition PointPosition { get; private set; }

        public BoxEntityPoint(Point point, PointPosition pointPosition) {
            this.Point = point;
            this.PointPosition = pointPosition;
        }
    }

    public enum RelationType { oneone, manymany, onemany, manyone, unknown }

    public class Relation {
        private BoxEntity sink;
        private BoxEntity source;
        private RelationType relationType;

        public BoxEntity Sink {
            get { return sink; }
        }

        public BoxEntity Source {
            get { return source; }
        }

        public RelationType RelationType {
            get { return this.relationType; }
        }

        public Relation(BoxEntity sink, BoxEntity source, RelationType relationType) {
            this.sink = sink;
            this.source = source;
            this.relationType = relationType;
        }

        public override bool Equals(Object obj) {
            // If parameter is null return false.
            if (obj == null) {
                return false;
            }

            // If parameter cannot be cast to Relation return false.
            var relation = obj as Relation;
            if (relation == null) {
                return false;
            }

            // Return true if the fields match:
            if (source == relation.source && sink == relation.sink && relationType == relation.relationType) return true;
            if (source == relation.sink && sink == relation.source) {
                if (relationType == relation.relationType && (relationType == RelationType.oneone || relationType == RelationType.manymany))
                    return true;
                if (relationType != relation.relationType) {
                    if (relationType == RelationType.manyone && relation.relationType == RelationType.onemany) return true;
                    if (relationType == RelationType.onemany && relation.relationType == RelationType.manyone) return true;
                }
            }
            return false;
        }

        public override int GetHashCode() {
            return source.GetHashCode() ^ sink.GetHashCode() ^ relationType.GetHashCode();
        }
    }
}