﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VisTool;
using System.ComponentModel;

namespace uVisStudio {
    public class StudioProperty {
        public Template Template; // Corresponds to Me
        public TplProperty TemplateProperty;
        public VisPropAtt attr = null;
        public VisComponent visComponent;
        public VisEngine.ControlAddress controlAddress;

        public StudioProperty(Template template, TplProperty tplProp, VisComponent visComponent) {
            this.Template = template;
            this.TemplateProperty = tplProp;
            this.visComponent = visComponent;
            if (visComponent != null)
                this.controlAddress = new VisEngine.ControlAddress(visComponent);
        }

        protected virtual void OnPropertyChanged(string propertyName) {
        }

        public virtual VisComponent GetVisComponent() {
            return this.controlAddress.GetControl();
        }

        public virtual string GetGUIType() {
            return Template.Info.Type.Name;
        }

        public virtual string GetFormula() {
            if (TemplateProperty == null || TemplateProperty.Formula == null) return "";
            return TemplateProperty.Formula;
        }

        public virtual string GetVisComponentValue() {
            if (this.attr.PropType == VisPropAtt.PropertyType.Eventhandler) return null;
            if (visComponent == null || TemplateProperty == null) return "No component selected";
            visComponent = this.controlAddress.GetControl();
            if (visComponent == null) return "No component selected";

            bool exists;
            dynamic value; // var x = typeof(VisClassAtt); 
            value = VisEngine.GetValue(visComponent, this.TemplateProperty.Name, out exists);
            if (value != null && value.GetType().IsArray) {
                if (value.Length == 0)
                    return "Zero elements";
                string result = CmpValue(value[0]);
                for (int i = 1; i < value.Length; i++) {
                    result += (", " + CmpValue(value[i]));
                }
                return result;
            }
            return CmpValue(value);
        }

        string CmpValue(object value) {
            try {
                VisComponent cmp;
                if (value is VisComponent) {
                    cmp = value as VisComponent;
                    return cmp.tpl.Name + "[" + cmp.Index() + ", last=" + cmp.Last() + "]";
                }
                if (value is BundleRef) {
                    BundleRef rf = value as BundleRef;
                    Bundle bdl = rf.bundle;
                    return "bundle[" + rf.index + ", last=" + (bdl.Instances.Count - 1) + "] is " + bdl.Target.tpl.Name + "[" + bdl.Target.Index() + "]";
                }

                if (value == null) return "null";
                if (value is DBNull) return "DBNull";
                if (value is string && String.IsNullOrWhiteSpace(value as string)) return "\"" + value + "\"";
            }
            catch { value = null; } // Indirect references may throw an exception, e.g. for deferred compile time errors, e.g. field doesn't exist.
            return Convert.ToString(value, VisEngine.codeCulture);
        }
    }

    //---------------------------------------------------------------------
    public class RowsProperty : StudioProperty {
        public RowsProperty(Template template, TplProperty property, VisComponent instance)
            : base(template, property, instance) { }

        public override string GetVisComponentValue() {
            string sql = Template.RowTpl.SQLstring;
            long count;
            if (Template.RowTpl.RowsExpr == null || Template.RowTpl.RowsExpr.RecordSet == null)
                return "Not a query: " + sql;
            count = Template.RowTpl.RowsExpr.RecordSet.Rows.Count;
            return count.ToString() + (count == 1 ? " row: " : " rows: ") + sql;
        }
    }

    public class NameProperty : StudioProperty {
        public NameProperty(Template template, TplProperty property, VisComponent instance)
            : base(template, property, instance) { }

        public override string GetVisComponentValue() {
            return "";
        }
    }

        public class TypeProperty : StudioProperty {
            public TypeProperty(Template template, TplProperty property, VisComponent instance)
                : base(template, property, instance) { }
            public override string GetVisComponentValue() {
                return "";
            }
        }
}
