﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using VisTool;

namespace uVisStudio.VisToolbox
{
    public partial class VisToolbox : UserControl
    {
        public List<VisToolboxGroup> ToolboxGroups { get; private set; }

        public VisToolbox()
        {
            this.InitializeComponent();
            //VisEngine.InitializeVisEngine(); // Ensure that VisEngine is initialized before looking for ComponentByName. SL 21-04-13
            ToolboxGroups = new List<VisToolboxGroup>();
            CreateToolboxGroups();
            ShowToolboxGroups();
        }

        private void CreateToolboxGroups() {
            var groupNames = new List<string>();

            foreach (var component in VisEngine.ComponentByName.Values) {
                if (component.ClassAttribute == null)
                    continue; // Virtual components such as VisComponent and ContainerComponent have ClassAttribute null.
                var name = component.ClassAttribute.ToolboxCategory;
                if (!groupNames.Contains(name)) {
                    var toolboxGroup = new VisToolboxGroup(name);
                    toolboxGroup.IsExpandedChanged += this.IsExpandedChanged;
                    toolboxGroup.IsExpanded = true;
                    ToolboxGroups.Add(toolboxGroup);
                    groupNames.Add(name);
                }
            }

            foreach (var component in VisEngine.ComponentByName.Values) {
                if (component.ClassAttribute == null) continue;
                var group = component.ClassAttribute.ToolboxCategory;
                foreach (var toolboxGroup in ToolboxGroups) {
                    if (toolboxGroup.nameLabel.Content.Equals(group)) {
                        if (component.ClassAttribute.ShowInToolbox)
                            toolboxGroup.AddToolboxItem(new VisToolboxItem(component));
                    }
                }
            }
        }

        private void IsExpandedChanged(object sender, EventArgs e)
        {
            this.ShowToolboxGroups();
        }

        private void ShowToolboxGroups()
        {
            List<VisToolboxItem> sortedList = new List<VisToolboxItem>();
            this.itemsPanel.Items.Clear();
            foreach (var toolboxGroup in this.ToolboxGroups)
            {
                //21-11-2012: Uncomment this code to show categories.

                //this.itemsPanel.Items.Add(toolboxGroup);
                //if (toolboxGroup.IsExpanded)
                //{
                    foreach (var item in toolboxGroup.ToolboxItems)
                    {
                        sortedList.Add(item);
                        
                    }
                //}                
            }
            //21-11-2012: Sorting the items.
            sortedList.Sort((a, b) => a.ComponentInfo.Type.Name.CompareTo(b.ComponentInfo.Type.Name));
            foreach (var item1 in sortedList)
                this.itemsPanel.Items.Add(item1);
        }
    }
}
