﻿using System.Drawing;
using System.IO;
using System.Diagnostics;
using System.Windows.Forms;
using System.Windows.Input;
using System.Windows.Media;
using VisTool;
using uVisStudio.VisDesigner;
using Brush = System.Windows.Media.Brush;
using Brushes = System.Windows.Media.Brushes;
using DataObject = System.Windows.DataObject;
using MouseEventArgs = System.Windows.Input.MouseEventArgs;
using UserControl = System.Windows.Controls.UserControl;

namespace uVisStudio.VisToolbox {
    public partial class VisToolboxItem : UserControl {
        public static VisToolboxItem SelectedTool;
        public static bool IsOneSelected;

        public ComponentInfo ComponentInfo { get; private set; }

        public VisToolboxItem() {
            this.InitializeComponent();
        }

        public VisToolboxItem(ComponentInfo componentInfo)
            : this() {
            this.ComponentInfo = componentInfo;
            this.itemLabel.Content = ComponentInfo.Type.Name;
            if (componentInfo.ClassAttribute == null) return;

            this.ToolTip = this.ComponentInfo.ClassAttribute.LongHelp == null ? this.ComponentInfo.ClassAttribute.ShortHelp : this.ComponentInfo.ClassAttribute.LongHelp;
            var icon = this.ComponentInfo.ClassAttribute.Icon;
            var iconPath = VisEngine.ProgramDirectory + "\\Icons\\" + icon;
            if (File.Exists(iconPath)) {
                var bitmap = new Bitmap(iconPath);
                this.itemIcon.Source = bitmap.GetImage();
                this.itemIcon.Stretch = Stretch.UniformToFill;
            }
        }


        public void ItemMouseUp(object sender, MouseButtonEventArgs e) {
            //if (StudioBase.Studio.IsMouseOver) {
            if (SelectedTool != this) { // New item selected
                if (SelectedTool != null)
                    SelectedTool.itemBorder.Background = ColorConstants.Background();
                SelectedTool = this;
                SelectedTool.itemBorder.Background = ColorConstants.Selected();
                IsOneSelected = true;
            } else if (IsOneSelected) {
                IsOneSelected = false; // Meaning multi draw.
                SelectedTool.itemBorder.Background = Brushes.LightGreen;
                SelectedTool.itemLabel.Content = SelectedTool.ComponentInfo.Type.Name + "  -  Multi";
            } else { // Was multi draw. Reset tool selection.
                SelectedTool.itemBorder.Background = ColorConstants.Background();
                SelectedTool.itemLabel.Content = SelectedTool.ComponentInfo.Type.Name;
                SelectedTool = null;
            }
            //    return;
            //}
            // Click outside studio. If VisForm selected, create a new VisForm
            //this.ReleaseMouseCapture();
            //System.Windows.Forms.Cursor.Current = System.Windows.Forms.Cursors.Default;
            //if (this.ComponentInfo.Type != typeof(VisForm))
            //    return;
            //System.Windows.Point wp = this.PointToScreen(e.GetPosition(this));
            //Point p = new Point((int)wp.X, (int)wp.Y);
            //var command = new TemplateDrawnCommand(this.ComponentInfo, null, p, 0, 0);
            //bool res = command.TryExecute();
            //if (res)
            //    UndoManager.AddCommand(command);
        }

        public static void Deselect() {
            IsOneSelected = false;
            if (SelectedTool == null) return;
            SelectedTool.itemBorder.Background = ColorConstants.Background();
            SelectedTool.itemLabel.Content = SelectedTool.ComponentInfo.Type.Name;
            SelectedTool = null;
        }
    }
}
